from .gui_app import TMSEEG_GUI

def main():
    """Entry point for the GUI application"""
    import tkinter as tk
    import sys
    import traceback
    from tkinter import messagebox
    
    def show_error(error_msg):
        """Display error in a simple tkinter window"""
        root = tk.Tk()
        root.withdraw()
        messagebox.showerror("Error", f"Failed to start GUI:\n{error_msg}")
        root.destroy()
    
    try:
        root = tk.Tk()
        root.title("TMS-EEG Analysis")
        app = TMSEEG_GUI(root)
        root.mainloop()
    except Exception as e:
        error_msg = f"{str(e)}\n\n{traceback.format_exc()}"
        show_error(error_msg)
        sys.exit(1)

if __name__ == "__main__":
    main()


from pathlib import Path

from _typeshed import Incomplete

has_icalabel_testing_data: Incomplete

def data_path(
    path: str | None = None,
    force_update: bool = False,
    update_path: bool = True,
    download: bool = True,
    verbose: Incomplete | None = None,
) -> Path:
    """ICA label testing data generated in conjunction with EEGLab.

    Parameters
    ----------
    path : None | str
        Location of where to look for the dataset.
        If None, the environment variable or config parameter is used.
        If it doesn’t exist, the “~/mne_data” directory is used.
        If the dataset is not found under the given path,
        the data will be automatically downloaded to the specified folder.
    force_update : bool
        Force update of the dataset even if a local copy exists.
    update_path : bool
        If True, set the MNE_DATASETS_FNIRSMOTORGROUP_PATH in
        mne-python config to the given path.
    download : bool
        If False and the dataset has not been downloaded yet,
        it will not be downloaded and the path will be returned
        as ``‘’`` (empty string). This is mostly used for debugging purposes
        and can be safely ignored by most users.

    verbose : bool | str | int | None
        Control verbosity of the logging output. If ``None``, use the default
        verbosity level. See the :ref:`logging documentation <tut-logging>` and
        :func:`mne.verbose` for details. Should only be passed as a keyword
        argument.

    Returns
    -------
    path : Path
        Path to dataset directory.
    """



import os.path as op

import pandas as pd
import pytest
from mne.datasets import testing
from mne.io import read_raw_edf
from mne.preprocessing import ICA
from mne_bids import BIDSPath, write_raw_bids

from tmseegpy.mne_icalabel.annotation import mark_component, write_components_tsv

subject_id = "01"
session_id = "01"
run = "01"
acq = "01"
task = "testing"
data_path = op.join(testing.data_path(), "EDF")
raw_fname = op.join(data_path, "test_reduced.edf")


@pytest.fixture
def tmp_bids_path(tmp_path):
    """Create a bids path."""
    bids_path = BIDSPath(
        subject=subject_id,
        session=session_id,
        run=run,
        acquisition=acq,
        task=task,
        root=tmp_path,
    )

    raw = read_raw_edf(raw_fname, verbose=False)
    bids_path = write_raw_bids(raw, bids_path, overwrite=True, verbose=False)
    return bids_path


@pytest.fixture(scope="session")
def ica():
    """Create an ICA object."""
    raw = read_raw_edf(raw_fname, preload=True)
    raw.filter(l_freq=1, h_freq=100)
    n_components = 5
    ica = ICA(n_components=n_components, method="picard")
    ica.fit(raw)
    return ica


def test_write_channels_tsv(ica: ICA, tmp_bids_path):
    """Test writing components to a channels.tsv file."""
    root = tmp_bids_path.root
    deriv_root = root / "derivatives" / "ICA"
    deriv_fname = BIDSPath(
        root=deriv_root,
        subject=subject_id,
        session=session_id,
        run=run,
        acquisition=acq,
        task=task,
        suffix="channels",
        extension=".tsv",
    )
    ica = ica.copy()
    ica.labels_["ecg"] = [0]

    write_components_tsv(ica, deriv_fname)

    assert deriv_fname.fpath.exists()
    expected_json = deriv_fname.copy().update(extension=".json")
    assert expected_json.fpath.exists()

    ch_tsv = pd.read_csv(deriv_fname, sep="\t")
    assert all(status == "good" for status in ch_tsv["status"][1:])
    assert ch_tsv["status"][0] == "bad"
    assert ch_tsv["ic_type"].values[0] == "ecg"


def test_mark_components(ica, tmp_bids_path):
    """Test marking components."""
    root = tmp_bids_path.root
    deriv_root = root / "derivatives" / "ICA"
    deriv_fname = BIDSPath(
        root=deriv_root,
        subject=subject_id,
        session=session_id,
        run=run,
        acquisition=acq,
        task=task,
        suffix="channels",
        extension=".tsv",
    )
    write_components_tsv(ica, deriv_fname)

    # mark components
    with pytest.raises(ValueError, match="not a valid label"):
        mark_component(0, deriv_fname, method="manual", label="heart", author="test")

    mark_component(0, deriv_fname, method="manual", label="heart beat", author="test")
    ch_tsv2 = pd.read_csv(deriv_fname, sep="\t")
    assert ch_tsv2["status"][0] == "bad"
    assert all(status == "good" for status in ch_tsv2["status"][1:])



from _typeshed import Incomplete
from mne import Info
from mne.preprocessing import ICA
from numpy.typing import NDArray

from ..utils._checks import _validate_ica as _validate_ica
from ..utils._docs import fill_doc as fill_doc

def get_topomaps(
    ica: ICA,
    picks: Incomplete | None = None,
    res: int = 64,
    image_interp: str = ...,
    border: float | str = ...,
    extrapolate: str = ...,
) -> dict[str, NDArray[float]]:
    """Generate an array of scalp topographies for the picked components.

    Parameters
    ----------
    ica : ICA
        MNE `~mne.preprocessing.ICA` decomposition.
    picks : int | list of int | slice | None
        Indices of the independent components (ICs) to select.
        If an integer, represents the index of the IC to pick.
        Multiple ICs can be selected using a list of int or a slice.
        The indices are 0-indexed, so ``picks=1`` will pick the second IC: ``ICA001``.
        ``None`` (default) will pick all independent components in the order fitted.
    res : int
        The resolution of the square topographic map (in pixels).
    image_interp : str
        The image interpolation to be used. All matplotlib options are
        accepted.
    border : float | 'mean'
        Value to extrapolate to on the topomap borders. If ``'mean'`` (default),
        then each extrapolated point has the average value of its neighbours.
    extrapolate : str
        Options:

        - ``'box'``
            Extrapolate to four points placed to form a square encompassing all
            data points, where each side of the square is three times the range
            of the data in the respective dimension.
        - ``'local'`` (default for MEG sensors)
            Extrapolate only to nearby points (approximately to points closer than
            median inter-electrode distance). This will also set the
            mask to be polygonal based on the convex hull of the sensors.
        - ``'head'`` (default for non-MEG sensors)
            Extrapolate out to the edges of the clipping circle. This will be on
            the head circle when the sensors are contained within the head circle,
            but it can extend beyond the head when sensors are plotted outside
            the head circle.

    Returns
    -------
    topomaps : dict of array of shape (n_components, n_pixels, n_pixels)
        Dictionary of ICs topographic maps for each channel type.
    """

def _get_topomap_array(
    data: NDArray[float],
    info: Info,
    res: int = 64,
    image_interp: str = ...,
    border: float | str = ...,
    extrapolate: str = ...,
) -> NDArray[float]:
    """Generate a scalp topographic map (n_pixels, n_pixels).

    Parameters
    ----------
    data : array of shape (n_channels,)
        The data points used to generate the topographic map.
    info : Info
        Instance of `mne.Info` with the montage associated with the ``(n_channels,)``
        points.
    res : int
        The resolution of the square topographic map (in pixels).
    image_interp : str
        The image interpolation to be used. All matplotlib options are
        accepted.
    border : float | 'mean'
        Value to extrapolate to on the topomap borders. If ``'mean'`` (default),
        then each extrapolated point has the average value of its neighbours.
    extrapolate : str
        Options:

        - ``'box'``
            Extrapolate to four points placed to form a square encompassing all
            data points, where each side of the square is three times the range
            of the data in the respective dimension.
        - ``'local'`` (default for MEG sensors)
            Extrapolate only to nearby points (approximately to points closer than
            median inter-electrode distance). This will also set the
            mask to be polygonal based on the convex hull of the sensors.
        - ``'head'`` (default for non-MEG sensors)
            Extrapolate out to the edges of the clipping circle. This will be on
            the head circle when the sensors are contained within the head circle,
            but it can extend beyond the head when sensors are plotted outside
            the head circle.

    Returns
    -------
    topomap : array of shape (n_pixels, n_pixels)
        Topographic map array.
    """



from mne import BaseEpochs
from mne.io import BaseRaw
from mne.preprocessing import ICA
from numpy.typing import NDArray

from ..utils._checks import _validate_inst_and_ica as _validate_inst_and_ica
from ._utils import _gdatav4 as _gdatav4
from ._utils import _mne_to_eeglab_locs as _mne_to_eeglab_locs
from ._utils import _next_power_of_2 as _next_power_of_2
from ._utils import _pol2cart as _pol2cart

def get_iclabel_features(inst: BaseRaw | BaseEpochs, ica: ICA):
    """Generate the features for ICLabel neural network.

    Parameters
    ----------
    inst : Raw | Epochs
        MNE Raw/Epoch instance with data array in Volts.
    ica : ICA
        MNE ICA decomposition.

    Returns
    -------
    topo : array of shape (32, 32, 1, n_components)
        The topoplot feature.
    psd : array of shape (1, 100, 1, n_components)
        The psd feature.
    autocorr : array of shape (1, 100, 1, n_components)
        The autocorrelations feature. Depending on the length of the
        raw data passed in, different methods of computing autocorrelation
        will be used. See :footcite:t:`PionTonachini2019` for details.

    References
    ----------
    .. footbibliography::
    """

def _retrieve_eeglab_icawinv(ica: ICA) -> tuple[NDArray[float], NDArray[float]]:
    """
    Retrieve 'icawinv' from an MNE ICA instance.

    Parameters
    ----------
    ica : ICA
        MNE ICA decomposition.

    Returns
    -------
    icawinv : array
    weights : array
    """

def _compute_ica_activations(inst: BaseRaw | BaseEpochs, ica: ICA) -> NDArray[float]:
    """Compute the ICA activations 'icaact' variable from an MNE ICA instance.

    Parameters
    ----------
    inst : Raw | Epoch
        MNE Raw/Epoch instance with data array in Volts.
    ica : ICA
        MNE ICA decomposition.

    Returns
    -------
    icaact : array
        raw: (n_components, n_samples)
        epoch: (n_components, n_samples, n_trials)

    Notes
    -----
    In EEGLAB, the ICA activations are computed after the data and the ICA
    decomposition are re-referenced to a common average, if the field 'EEG.ref'
    is different from 'averef'. The EEGLAB sample dataset's field 'EEG.ref' is
    set to 'common', thus triggering the re-referencing with 'pop_reref' which
    seems to be buggy and breaks the ICA solution. After 'pop_reref' is called,
    the relation 'inv(EEG.icaweights * EEG.icasphere) = EEG.icawinv' is not
    respected anymore.

    Additionally, 'pop_reref' changes the field 'EEG.ref' to 'average'. It is
    assumed that 'common', 'average' and 'averef' are all denoting a common
    average reference.
    """

def _eeg_topoplot(
    inst: BaseRaw | BaseEpochs, icawinv: NDArray[float], picks: list[str]
) -> NDArray[float]:
    """Topoplot feature."""

def _topoplotFast(
    values: NDArray[float], rd: NDArray[float], th: NDArray[float]
) -> NDArray[float]:
    """Implement topoplotFast.m from MATLAB. Each topographic map is a 32x32 images."""

def _eeg_rpsd(
    inst: BaseRaw | BaseEpochs, ica: ICA, icaact: NDArray[float]
) -> NDArray[float]:
    """PSD feature."""

def _eeg_rpsd_constants(
    inst: BaseRaw | BaseEpochs, ica: ICA
) -> tuple[int, int, int, int, NDArray[int], NDArray[float], NDArray[int]]:
    """Compute the constants before ``randperm`` is used to compute the subset."""

def _eeg_rpsd_compute_psdmed(
    inst: BaseRaw | BaseEpochs,
    icaact: NDArray[float],
    ncomp: int,
    nfreqs: int,
    n_points: int,
    nyquist: int,
    index: NDArray[int],
    window: NDArray[float],
    subset: NDArray[int],
) -> NDArray[float]:
    """Compute the variable 'psdmed', annotated as windowed spectrums."""

def _eeg_rpsd_format(psd: NDArray[float]) -> NDArray[float]:
    """Apply the formatting steps after 'eeg_rpsd.m'."""

def _eeg_autocorr_welch(
    raw: BaseRaw, ica: ICA, icaact: NDArray[float]
) -> NDArray[float]:
    """Autocorrelation feature applied on raw object with at least 5 * fs samples.

    MATLAB: 'eeg_autocorr_welch.m'.
    """

def _eeg_autocorr(raw: BaseRaw, ica: ICA, icaact: NDArray[float]) -> NDArray[float]:
    """Autocorr applied on raw object without enough samples for eeg_autocorr_welch.

    MATLAB: 'eeg_autocorr.m'.
    """

def _eeg_autocorr_fftw(
    epochs: BaseEpochs, ica: ICA, icaact: NDArray[float]
) -> NDArray[float]:
    """Autocorrelation feature applied on epoch object.

    MATLAB: 'eeg_autocorr_fftw.m'.
    """

def _resample(ac: NDArray[float], fs: int | float) -> NDArray[float]:
    """Resample the autocorrelation feature.

    The comment in EEGLAB is:
        resample to 1 second at 100 samples/sec

    Which translates by: the output array must be of shape (n_comp, 101), thus
    the resampling up variable is set to 100, and down variable must respect:
        100 < ac.T.shape[0] * 100 / down <= 101
    If the instance sampling frequency is an integer, then down is equal to the
    sampling frequency.

    Parameters
    ----------
    ac : array
        Array of shape (n_comp, samples).
    fs : int | float
        Sampling frequency of the MNE instance.
    """



from mne.preprocessing import ICA

def label_ica_components(inst, ica: ICA, show: bool = True, block: bool = False):
    """Launch the IC labelling GUI.

    Parameters
    ----------
    inst : Raw | Epochs
        `~mne.io.Raw` or `~mne.Epochs` instance used to fit the `~mne.preprocessing.ICA`
        decomposition.
    ica : ICA
        The ICA object fitted on ``inst``.
    show : bool
        Show the GUI if True.
    block : bool
        Whether to halt program execution until the figure is closed.

    Returns
    -------
    gui : instance of ICAComponentLabeler
        The graphical user interface (GUI) window.
    """



# run.py
import os
from pathlib import Path
import numpy as np
import matplotlib.pyplot as plt
from .preproc import TMSEEGPreprocessor
from .pcist import PCIst
from .preproc_vis import plot_raw_segments, plot_epochs_grid
from .validate_tep import plot_tep_analysis, generate_validation_summary
import mne
import time
from .neurone_loader import Recording
import argparse

mne.viz.use_browser_backend("matplotlib")
plt.rcParams['figure.figsize'] = [8, 6]


def generate_preproc_stats(processor, session_name, output_dir):
    """
    Generate simplified preprocessing quality control report.
    
    Args:
        processor: TMSEEGPreprocessor object
        session_name: Name of the current session
        output_dir: Directory to save the output file
    """
    from pathlib import Path
    import datetime
    import numpy as np
    
    output_file = Path(output_dir) / f"preproc_stats_{session_name}_{datetime.datetime.now().strftime('%Y%m%d_%H%M%S')}.txt"
    
    with open(output_file, 'w') as f:
        # Header with key session info
        f.write(f"TMS-EEG Preprocessing Report: {session_name}\n")
        f.write("=" * 50 + "\n\n")
        
        # Recording parameters
        f.write("Recording Parameters\n")
        f.write("-" * 20 + "\n")
        f.write(f"Duration: {processor.raw.times[-1]:.1f} seconds\n")
        f.write(f"Sampling rate: {processor.raw.info['sfreq']} → {processor.ds_sfreq} Hz\n")
        f.write(f"Channels: {len(processor.raw.ch_names)}\n\n")
        
        '''# TMS pulse information
        events = processor._get_events()
        n_events = len(events)
        f.write("TMS Pulses\n")
        f.write("-" * 20 + "\n")
        f.write(f"Total pulses: {n_events}\n")
        if n_events > 1:
            intervals = np.diff([event[0] for event in events]) / processor.raw.info['sfreq']
            f.write(f"Mean interval: {np.mean(intervals):.3f} s (±{np.std(intervals):.3f})\n\n")'''
        
        # Data quality summary
        f.write("Data Quality Metrics\n")
        f.write("-" * 20 + "\n")
        
        # Channel quality
        n_bad_channels = len(getattr(processor, 'bad_channels', []))
        channel_retention = (len(processor.raw.ch_names) - n_bad_channels) / len(processor.raw.ch_names) * 100
        f.write(f"Channel retention: {channel_retention:.1f}%")
        if hasattr(processor, 'bad_channels') and processor.bad_channels:
            f.write(f" (Removed: {', '.join(processor.bad_channels)})")
        f.write("\n")
        
        # Epoch quality
        if hasattr(processor, 'epochs'):
            n_bad_epochs = len(getattr(processor, 'bad_epochs', []))
            epoch_retention = (len(processor.epochs) - n_bad_epochs) / len(processor.epochs) * 100
            f.write(f"Epoch retention: {epoch_retention:.1f}%")
            if n_bad_epochs > 0:
                f.write(f" ({n_bad_epochs} epochs removed)")
            f.write("\n")
        
        # Artifact removal summary
        f.write("\nArtifact Removal\n")
        f.write("-" * 20 + "\n")
        
        # ICA components
        if hasattr(processor, 'muscle_components'):
            f.write(f"Muscle components removed: {len(processor.muscle_components)}\n")
        if hasattr(processor, 'excluded_components'):
            f.write(f"Other artifacts removed: {len(processor.excluded_components)}\n")
        
        # Overall quality assessment
        f.write("\nQuality Assessment\n")
        f.write("-" * 20 + "\n")
        
        # Calculate simplified quality score
        quality_score = np.mean([
            channel_retention / 100,
            epoch_retention / 100 if hasattr(processor, 'epochs') else 1.0
        ])
        
        f.write(f"Overall quality score: {quality_score*100:.1f}%\n")
        
        # Add focused warnings if needed
        if quality_score < 0.7:
            f.write("\nWarnings:\n")
            if channel_retention < 80:
                f.write("• High number of channels removed\n")
            if hasattr(processor, 'epochs') and epoch_retention < 80:
                f.write("• High number of epochs removed\n")
    
    return output_file

def generate_research_stats(pcist_values, pcist_objects, details_list, session_names, output_dir):
    """
    Generate detailed research statistics for PCIst measurements, including both individual
    session statistics and pooled analysis across all sessions.
    
    Args:
        pcist_values: List of PCIst values for all sessions
        pcist_objects: List of PCIst objects containing the analyses
        details_list: List of dictionaries containing PCIst calculation details for each session
        session_names: List of session names
        output_dir: Directory to save the output file
    """
    import numpy as np
    from pathlib import Path
    from scipy import stats
    import datetime
    
    output_file = Path(output_dir) / f"pcist_research_stats_{datetime.datetime.now().strftime('%Y%m%d_%H%M%S')}.txt"
    
    with open(output_file, 'w') as f:
        # Header
        f.write("PCIst Research Statistics Report\n")
        f.write("=" * 50 + "\n\n")
        
        # 1. Pooled Statistics
        f.write("1. POOLED STATISTICS ACROSS ALL SESSIONS\n")
        f.write("-" * 40 + "\n")
        
        # Basic descriptive statistics
        pcist_array = np.array(pcist_values)
        f.write(f"Number of sessions: {len(pcist_values)}\n")
        f.write(f"Mean PCIst: {np.mean(pcist_array):.4f}\n")
        f.write(f"Median PCIst: {np.median(pcist_array):.4f}\n")
        f.write(f"Standard deviation: {np.std(pcist_array):.4f}\n")
        f.write(f"Coefficient of variation: {(np.std(pcist_array)/np.mean(pcist_array))*100:.2f}%\n")
        f.write(f"Range: {np.min(pcist_array):.4f} - {np.max(pcist_array):.4f}\n")
        
        # Distribution statistics
        f.write("\nDistribution Statistics:\n")
        f.write(f"Skewness: {stats.skew(pcist_array):.4f}\n")
        f.write(f"Kurtosis: {stats.kurtosis(pcist_array):.4f}\n")
        shapiro_stat, shapiro_p = stats.shapiro(pcist_array)
        f.write(f"Shapiro-Wilk test (normality): W={shapiro_stat:.4f}, p={shapiro_p:.4f}\n")
        
        # Quartile statistics
        q1, q2, q3 = np.percentile(pcist_array, [25, 50, 75])
        iqr = q3 - q1
        f.write(f"\nQuartile Statistics:\n")
        f.write(f"Q1 (25th percentile): {q1:.4f}\n")
        f.write(f"Q2 (median): {q2:.4f}\n")
        f.write(f"Q3 (75th percentile): {q3:.4f}\n")
        f.write(f"IQR: {iqr:.4f}\n")
        
        # Outlier detection
        lower_bound = q1 - 1.5 * iqr
        upper_bound = q3 + 1.5 * iqr
        outliers = pcist_array[(pcist_array < lower_bound) | (pcist_array > upper_bound)]
        outlier_sessions = [session_names[i] for i, val in enumerate(pcist_array) 
                          if val < lower_bound or val > upper_bound]
        
        f.write("\nOutlier Analysis:\n")
        f.write(f"Lower bound: {lower_bound:.4f}\n")
        f.write(f"Upper bound: {upper_bound:.4f}\n")
        f.write(f"Number of outliers: {len(outliers)}\n")
        if outliers.size > 0:
            f.write("Outlier sessions:\n")
            for session, value in zip(outlier_sessions, outliers):
                f.write(f"  {session}: {value:.4f}\n")
        
        # 2. Session-by-Session Analysis
        f.write("\n2. INDIVIDUAL SESSION STATISTICS\n")
        f.write("-" * 40 + "\n")
        
        for i, (pcist_obj, details, session_name) in enumerate(zip(pcist_objects, details_list, session_names)):
            f.write(f"\nSession: {session_name}\n")
            f.write(f"PCIst value: {pcist_values[i]:.4f}\n")
            
            # SVD and Component Analysis
            if 'signal_shape' in details:
                f.write(f"\nSignal Analysis:\n")
                f.write(f"Signal shape: {details['signal_shape']}\n")
            
            if 'singular_values' in details:
                sv = details['singular_values']
                f.write(f"Number of components: {len(sv)}\n")
                f.write(f"Non-zero components: {np.sum(sv > 1e-10)}\n")
                if 'variance_explained' in details:
                    f.write(f"Variance explained: {details['variance_explained']:.2f}%\n")
            
            # SNR Information
            if 'snr_values' in details:
                snr_values = details['snr_values']
                f.write("\nSNR Statistics:\n")
                f.write(f"Mean SNR: {np.mean(snr_values):.4f}\n")
                f.write(f"Max SNR: {np.max(snr_values):.4f}\n")
                if 'parameters' in details and 'min_snr' in details['parameters']:
                    f.write(f"Channels above SNR threshold: {np.sum(snr_values > details['parameters']['min_snr'])}\n")
            
            # NST Analysis
            if 'nst_base_range' in details and 'nst_resp_range' in details:
                f.write("\nNST Analysis:\n")
                f.write(f"NST baseline range: {details['nst_base_range'][0]:.3f} to {details['nst_base_range'][1]:.3f}\n")
                f.write(f"NST response range: {details['nst_resp_range'][0]:.3f} to {details['nst_resp_range'][1]:.3f}\n")
            
            if 'dnst_values' in details:
                f.write(f"Mean dNST: {np.mean(details['dnst_values']):.4f}\n")
                f.write(f"Max dNST: {np.max(details['dnst_values']):.4f}\n")
            
            # Parameters used
            if 'parameters' in details:
                f.write("\nAnalysis Parameters:\n")
                for param, value in details['parameters'].items():
                    f.write(f"{param}: {value}\n")
            
        # 3. Quality Control Summary
        f.write("\n3. QUALITY CONTROL SUMMARY\n")
        f.write("-" * 40 + "\n")
        
        # Overall quality metrics
        mean_snr = np.mean([np.mean(d.get('snr_values', [0])) for d in details_list])
        mean_var_explained = np.mean([d.get('variance_explained', 0) for d in details_list])
        
        f.write(f"Mean SNR across sessions: {mean_snr:.4f}\n")
        f.write(f"Mean variance explained: {mean_var_explained:.2f}%\n")
        
        # Quality assessment
        quality_threshold = 0.5
        quality_metrics = [
            mean_snr / 10,
            mean_var_explained / 100,
            1 - (len(outliers) / len(pcist_values))
        ]
        overall_quality = np.mean(quality_metrics)
        
        f.write(f"\nOverall Quality Assessment: {overall_quality:.4f}\n")
        if overall_quality < quality_threshold:
            f.write("WARNING: Quality metrics below threshold - review data carefully\n")
        
    return output_file


def process_subjects(args):
    import builtins
    setattr(builtins, 'STOP_PROCESSING', False)
    def check_stop():
        if getattr(builtins, 'STOP_PROCESSING', False):
            print("\nProcessing stopped by user")
            return True
        return False
    data_dir = Path(args.data_dir)
    # Specific paths for the subject
    TMS_DATA_PATH = data_dir / 'TMSEEG'
    # Verify paths exist
    required_paths = {
        'Data Directory': data_dir,
        'TMS Data': TMS_DATA_PATH,
    }
    for name, path in required_paths.items():
        if not path.exists():
            print(f"WARNING: {name} not found at: {path}")
        else:
            print(f"✓ Found {name} at: {path}")
    
    # Store PCI values
    subject_pcist_values = []
    pcist_objects = []
    pcist_details = []
    session_names = []

    # Load the raw data using the new loader
    from .dataloader import TMSEEGLoader
    loader = TMSEEGLoader(
        data_path=TMS_DATA_PATH,
        format=args.data_format,
        substitute_zero_events_with=args.substitute_zero_events_with,
        eeglab_montage_units=args.eeglab_montage_units,
        verbose=True
    )

    raw_list = loader.load_data()
    session_info = loader.get_session_info()

    np.random.seed(args.random_seed)
    baseline_start_sec = args.baseline_start / 1000.0
    baseline_end_sec = args.baseline_end / 1000.0

    # Loop through the loaded raw data
    for n, raw in enumerate(raw_list):
        if check_stop():
            return []
        
        session_name = session_info[n]['name']
        print(f"\nProcessing Session {n}: {session_name}")

        if check_stop(): return []

        # Process session...
        try:
            # First try user-specified channel
            events = mne.find_events(raw, stim_channel=args.stim_channel)
        except ValueError:
            try:
                # If no stim channel found, try to get events from annotations
                print("No stim channel found, looking for events in annotations...")
                
                # Get unique annotation descriptions
                unique_descriptions = set(raw.annotations.description)
                print(f"Found annotation types: {unique_descriptions}")
                
                # For TMS-EEG we typically want 'Stimulation' or similar annotations
                tms_annotations = ['Stimulation', 'TMS', 'R128', 'Response']
                
                # Create mapping for event IDs
                event_id = {}
                for desc in unique_descriptions:
                    # Look for TMS-related annotations
                    if any(tms_str.lower() in desc.lower() for tms_str in tms_annotations):
                        event_id[desc] = args.substitute_zero_events_with
                
                if not event_id:
                    print("No TMS-related annotations found. Using all annotations...")
                    # If no TMS annotations found, use all annotations
                    for i, desc in enumerate(unique_descriptions, 1):
                        event_id[desc] = i
                
                print(f"Using event mapping: {event_id}")
                
                # Get events from annotations
                events, _ = mne.events_from_annotations(raw, event_id=event_id)
                
                if len(events) == 0:
                    raise ValueError("No events found in annotations")
                    
                print(f"Found {len(events)} events from annotations")
                
            except Exception as e:
                # If both methods fail, try common stim channels
                print(f"Could not get events from annotations: {str(e)}")
                print("Trying common stim channel names...")
                
                stim_channels = mne.pick_types(raw.info, stim=True, exclude=[])
                if len(stim_channels) > 0:
                    stim_ch_name = raw.ch_names[stim_channels[0]]
                    print(f"Using detected stim channel: {stim_ch_name}")
                    events = mne.find_events(raw, stim_channel=stim_ch_name)
                else:
                    common_stim_names = ['STI 014', 'STIM', 'STI101', 'trigger', 'STI 001']
                    found = False
                    for ch_name in common_stim_names:
                        if ch_name in raw.ch_names:
                            print(f"Using stim channel: {ch_name}")
                            events = mne.find_events(raw, stim_channel=ch_name)
                            found = True
                            break
                    if not found:
                        raise ValueError(f"Could not find events in data. Available channels: {raw.ch_names}")

        print(f"Found {len(events)} events")

        print(f"Found {len(events)} events")
        annotations = mne.annotations_from_events(
            events=events, 
            sfreq=raw.info['sfreq'],
            event_desc={args.substitute_zero_events_with: 'Stimulation'}
        )
        raw.set_annotations(annotations)

        # Drop unnecessary channels
        channels_to_drop = []
        if 'EMG1' in raw.ch_names:
            channels_to_drop.append('EMG1')
        if channels_to_drop:
            print(f"Dropping channels: {channels_to_drop}")
            raw.drop_channels(channels_to_drop)
        #if args.plot_preproc:
           # plot_raw_segments(raw, args.output_dir, step_name='raw',)

        # Preprocessing
        processor = TMSEEGPreprocessor(raw, ds_sfreq=args.ds_sfreq)
        print("\nRemoving TMS artifact and muscle peaks...")
        if check_stop(): return []
        processor.remove_tms_artifact(cut_times_tms=(-2, 10))  # Step 8

        print("\nInterpolating TMS artifact...")
        processor.interpolate_tms_artifact(method='cubic', 
                                        interp_window=1.0,  # 1ms window for initial interpolation
                                        cut_times_tms=(-2, 10))  # Step 9

        #if args.plot_preproc:
           # plot_raw_segments(raw, args.output_dir, step_name='raw_i',)
        #processor.fix_tms_artifact(window=(args.fix_artifact_window_start, args.fix_artifact_window_end))
        print("\nFiltering raw eeg data...")
        if check_stop(): return []
        processor.filter_raw(l_freq=args.l_freq, h_freq=args.h_freq, notch_freq=args.notch_freq, notch_width=args.notch_width, plot_psd=False)

        #if args.plot_preproc:
          #  plot_raw_segments(raw, args.output_dir, step_name='raw_f',)

        print("\nCreating epochs...")
        processor.create_epochs(tmin=args.epochs_tmin, tmax=args.epochs_tmax, baseline=None, amplitude_threshold=args.amplitude_threshold)
        epochs = processor.epochs

        print("\nRemoving bad channels...")
        processor.remove_bad_channels(threshold=args.bad_channels_threshold)

        print("\nRemoving bad epochs...")
        processor.remove_bad_epochs(threshold=args.bad_epochs_threshold)
        if args.plot_preproc:
            plot_epochs_grid(epochs, args.output_dir, session_name=session_name, step_name='epochs')

        print("\nSetting average reference...")
        processor.set_average_reference()

        print("\nRunning first ICA...")
        if check_stop(): return []
        if args.plot_preproc:
            plot_components=True
        else:
            plot_components=False
        processor.run_ica(output_dir=args.output_dir, session_name=session_name, method=args.ica_method, tms_muscle_thresh=args.tms_muscle_thresh, plot_components=plot_components)
        if args.plot_preproc:
            plot_epochs_grid(epochs, args.output_dir, session_name=session_name, step_name='ica1')
        if args.clean_muscle_artifacts:
            print("\nCleaning muscle artifacts...")
            if check_stop(): return []
            processor.clean_muscle_artifacts(
                muscle_window=(args.muscle_window_start, args.muscle_window_end),
                threshold_factor=args.threshold_factor,
                n_components=args.n_components,
                verbose=True
            )
        if args.plot_preproc:
            plot_epochs_grid(epochs, args.output_dir, session_name=session_name, step_name='clean')
        if not args.skip_second_artifact_removal:
            print("\nExtending TMS artifact removal window...")
            processor.remove_tms_artifact(cut_times_tms=(-2, 15))  
            
            print("\nInterpolating extended TMS artifact...")
            processor.interpolate_tms_artifact(method='cubic',
                                            interp_window=5.0,  
                                            cut_times_tms=(-2, 15))
            
        # https://mne.tools/mne-icalabel/stable/generated/examples/00_iclabel.html#sphx-glr-generated-examples-00-iclabel-py

        if not args.no_second_ICA:
            print("\nRunning second ICA...")
            if check_stop(): return []
            processor.run_second_ica(method=args.second_ica_method, exclude_labels=["eye blink", "heart beat", "muscle artifact", "channel noise", "line noise"])

        if args.apply_ssp:    
            print("\nApplying SSP...")
            processor.apply_ssp(n_eeg=args.ssp_n_eeg)

        print("\nApplying baseline correction...")
        processor.apply_baseline_correction(baseline=(baseline_start_sec, baseline_end_sec))
        if args.plot_preproc:
            plot_epochs_grid(epochs, args.output_dir, session_name=session_name, step_name='ica2')
        if args.apply_csd:
            print("\nApplying CSD transformation...")
            processor.apply_csd(lambda2=args.lambda2, stiffness=args.stiffness)
    
        print(f"\nDownsampling to {processor.ds_sfreq} Hz")
        processor.downsample()
        if args.plot_preproc:
            plot_epochs_grid(epochs, args.output_dir, session_name, step_name='final')

        epochs = processor.epochs
        epochs.save(Path(args.output_dir) / f"{session_name}_preproc-epo.fif")

        if args.validate_teps:
            try:
                print("\nAnalyzing TEPs...")
                if check_stop(): return []
                
                # Calculate evoked response
                evoked = epochs.average()
                
                # Basic validation checks
                if not evoked.times.size:
                    raise ValueError("No time points in evoked data")
                if not evoked.data.any():
                    raise ValueError("No data in evoked response")
                    
                print(f"Processing TEPs from {evoked.times[0]:.3f}s to {evoked.times[-1]:.3f}s")
                
                # Run TEP analysis
                components = plot_tep_analysis(
                    evoked=evoked,
                    output_dir=args.output_dir,
                    session_name=session_name,
                    prominence=args.prominence
                )
                
                # Check if any components were found
                if not components:
                    print("Warning: No TEP components were detected")
                else:
                    print(f"Found {len(components)} TEP components")
                    
                    if args.save_validation:
                        generate_validation_summary(
                            components,
                            args.output_dir,
                            session_name
                        )
                        print("TEP validation summary saved")
                        
            except Exception as e:
                print(f"Warning: TEP analysis failed: {str(e)}")
                import traceback
                print("Detailed error:")
                print(traceback.format_exc())
                components = None
        
        # Final quality check
        fig = processor.plot_evoked_response(ylim={'eeg': [-2, 2]}, xlim=(-0.1, 0.4), title="Final Evoked Response", show=args.show_evoked)
        fig.savefig(f"{args.output_dir}/evoked_{session_name}.png")  
        plt.close(fig)
        
        recording_id = f"session_{n}"

        # PCIst analysis
        pcist = PCIst(epochs)
        par = {
            'baseline_window': (args.baseline_start, args.baseline_end),
            'response_window': (args.response_start, args.response_end),
            'k': args.k,
            'min_snr': args.min_snr,
            'max_var': args.max_var,
            'embed': args.embed,
            'n_steps': args.n_steps
        }
        value, details = pcist.calc_PCIst(**par, return_details=True)
        fig = pcist.plot_analysis(details, session_name=session_name)
        print(f"PCI: {value}")
        subject_pcist_values.append(value)
        fig.savefig(f"{args.output_dir}/pcist_{session_name}.png")  
        plt.close(fig)

        pcist_objects.append(pcist)
        pcist_details.append(details)
        session_names.append(session_name)

    
    #if args.preproc_qc:
       # preproc_stats_file = generate_preproc_stats(
           # processor,
           # session_name,
           # args.output_dir
       #)
       # print(f"Preprocessing statistics saved to: {preproc_stats_file}")

    if args.research:
        output_file = generate_research_stats(
            subject_pcist_values,
            pcist_objects,
            pcist_details,
            session_names,
            args.output_dir
        )
        print(f"Research statistics saved to: {output_file}")
 
    return subject_pcist_values

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Process EEG data.')
    parser.add_argument('--data_dir', type=str, default=str(Path.cwd() / 'data'), 
                        help='Path to the data directory (default: ./data)')
    parser.add_argument('--output_dir', type=str, default=str(Path.cwd() / 'output'), 
                        help='Path to the output directory (default: ./output)')
    parser.add_argument('--data_format', type=str, default='neurone',
                   choices=['neurone', 'brainvision', 'edf', 'cnt', 'eeglab', 'auto'],
                   help='Format of input data (default: neurone)')
    parser.add_argument('--eeglab_montage_units', type=str, default='auto',
                   help='Units for EEGLAB channel positions (default: auto)')
    parser.add_argument('--stim_channel', type=str, default='STI 014',
                    help='Name of the stimulus channel (default: STI 014)')
    parser.add_argument('--plot_preproc', action='store_true',
                    help='Enable muscle artifact cleaning (default: False)')
    parser.add_argument('--random_seed', type=int, default=42,
                        help='Random seed for reproducibility (default: 42)')
    parser.add_argument('--substitute_zero_events_with', type=int, default=10,
                        help='Value to substitute zero events with (default: 10)')
    parser.add_argument('--ds_sfreq', type=float, default=725,
                        help='Downsampling frequency (default: 725)')
    # Trying to match TESA
    parser.add_argument('--initial_window_start', type=float, default=-2,
                    help='Initial TMS artifact window start (TESA default: -2)')
    parser.add_argument('--initial_window_end', type=float, default=10,
                        help='Initial TMS artifact window end (TESA default: 10)')
    parser.add_argument('--extended_window_start', type=float, default=-2,
                        help='Extended TMS artifact window start (TESA default: -2)')
    parser.add_argument('--extended_window_end', type=float, default=15,
                        help='Extended TMS artifact window end (TESA default: 15)')
    parser.add_argument('--initial_interp_window', type=float, default=1.0,
                        help='Initial interpolation window (TESA default: 1.0)')
    parser.add_argument('--extended_interp_window', type=float, default=5.0,
                        help='Extended interpolation window (TESA default: 5.0)')
    parser.add_argument('--interpolation_method', type=str, default='cubic',
                        choices=['cubic'],
                        help='Interpolation method (TESA requires cubic)')
    parser.add_argument('--skip_second_artifact_removal', action='store_true',
                    help='Skip the second stage of TMS artifact removal')
    parser.add_argument('--l_freq', type=float, default=0.1,
                        help='Lower frequency for filtering (default: 0.1)')
    parser.add_argument('--h_freq', type=float, default=45,
                        help='Upper frequency for filtering (default: 45)')
    parser.add_argument('--notch_freq', type=float, default=50,
                        help='Notch filter frequency (default: 50)')
    parser.add_argument('--notch_width', type=float, default=2,
                        help='Notch filter width (default: 2)')
    parser.add_argument('--epochs_tmin', type=float, default=-0.41,
                        help='Start time for epochs (default: -0.41)')
    parser.add_argument('--epochs_tmax', type=float, default=0.41,
                        help='End time for epochs (default: 0.41)')
    parser.add_argument('--bad_channels_threshold', type=float, default=3,
                        help='Threshold for removing bad channels (default: 3)')
    parser.add_argument('--bad_epochs_threshold', type=float, default=3,
                        help='Threshold for removing bad epochs (default: 3)')
    parser.add_argument('--ica_method', type=str, default='fastica',
                        help='ICA method (default: fastica)')
    parser.add_argument('--tms_muscle_thresh', type=float, default=3.0,
                        help='Threshold for TMS muscle artifact (default: 3.0)')
    parser.add_argument('--clean_muscle_artifacts', action='store_true',
                    help='Enable muscle artifact cleaning (default: False)')
    parser.add_argument('--muscle_window_start', type=float, default=0.005,
                    help='Start time for muscle artifact window (default: 0.005)')
    parser.add_argument('--muscle_window_end', type=float, default=0.030,
                    help='End time for muscle artifact window (default: 0.030)')
    parser.add_argument('--threshold_factor', type=float, default=1.0,
                    help='Threshold factor for muscle artifact cleaning (default: 1.0)')
    parser.add_argument('--n_components', type=int, default=5,
                    help='Number of components for muscle artifact cleaning (default: 5)')
    parser.add_argument('--no_second_ICA', action='store_true',
                    help='Disable seconds ICA using ICA_label (default: False)')
    parser.add_argument('--second_ica_method', type=str, default='infomax',
                        help='Second ICA method (default: infomax)')
    parser.add_argument('--apply_ssp', action='store_true',
                    help='Apply SSP (default: False)')
    parser.add_argument('--ssp_n_eeg', type=int, default=2,
                        help='Number of EEG components for SSP (default: 2)')
    parser.add_argument('--apply_csd', action='store_true',
                    help='Apply CSD transformation (default: True)')
    parser.add_argument('--lambda2', type=float, default=1e-3,
                    help='Lambda2 parameter for CSD transformation (default: 1e-5)')
    parser.add_argument('--stiffness', type=int, default=4,
                    help='Stiffness parameter for CSD transformation (default: 4)')
    parser.add_argument('--show_evoked', action='store_true',
                    help='Display the evoked plot with TEPs (default: False)')
    parser.add_argument('--validate_teps', action='store_true',
                help='Perform TEP validation against established criteria')
    parser.add_argument('--save_validation', action='store_true',
                help='Save TEP validation summary (default: False)')
    parser.add_argument('--prominence', type=float, default=0.01,
                    help='Minimum prominence for peak detection (default: 0.01)')
    parser.add_argument('--baseline_start', type=int, default=-400,
                        help='Start time for baseline in ms (default: -400)')
    parser.add_argument('--baseline_end', type=int, default=-50,
                        help='End time for baseline in ms (default: -50)')
    parser.add_argument('--response_start', type=int, default=0,
                        help='Start of response window in ms (default: 0)')
    parser.add_argument('--response_end', type=int, default=299,
                        help='End of response window in ms (default: 299)')
    parser.add_argument('--amplitude_threshold', type=float, default=300.0,
                    help='Threshold for epoch rejection based on peak-to-peak amplitude in µV (default: 300.0)')
    parser.add_argument('--k', type=float, default=1.2,
                        help='PCIst parameter k (default: 1.2)')
    parser.add_argument('--min_snr', type=float, default=1.1,
                        help='PCIst parameter min_snr (default: 1.1)')
    parser.add_argument('--max_var', type=float, default=99.0,
                        help='PCIst parameter max_var (default: 99.0)')
    parser.add_argument('--embed', action='store_true',
                        help='PCIst parameter embed (default: False)')
    parser.add_argument('--n_steps', type=int, default=100,
                        help='PCIst parameter n_steps (default: 100)')
    parser.add_argument('--pre_window_start', type=int, default=-400,
                        help='Start of the pre-TMS window in ms (default: -400)')
    parser.add_argument('--pre_window_end', type=int, default=-50,
                        help='End of the pre-TMS window in ms (default: -50)')
    parser.add_argument('--post_window_start', type=int, default=0,
                        help='Start of the post-TMS window in ms (default: 0)')
    parser.add_argument('--post_window_end', type=int, default=300,
                        help='End of the post-TMS window in ms (default: 300)')
    parser.add_argument('--research', action='store_true',
                    help='Output summary statistics of measurements (default: False)')
    args = parser.parse_args()

    pcists = process_subjects(args)
    print(f"PCIst values: {pcists}")


import torch
import torch.nn as nn
from _typeshed import Incomplete
from numpy.typing import ArrayLike, NDArray

from .utils import _format_input as _format_input

class _ICLabelNetImg(nn.Module):
    conv1: Incomplete
    relu1: Incomplete
    conv2: Incomplete
    relu2: Incomplete
    conv3: Incomplete
    relu3: Incomplete
    sequential: Incomplete

    def __init__(self) -> None: ...
    def forward(self, x): ...

class _ICLabelNetPSDS(nn.Module):
    conv1: Incomplete
    relu1: Incomplete
    conv2: Incomplete
    relu2: Incomplete
    conv3: Incomplete
    relu3: Incomplete
    sequential: Incomplete

    def __init__(self) -> None: ...
    def forward(self, x): ...

class _ICLabelNetAutocorr(nn.Module):
    conv1: Incomplete
    relu1: Incomplete
    conv2: Incomplete
    relu2: Incomplete
    conv3: Incomplete
    relu3: Incomplete
    sequential: Incomplete

    def __init__(self) -> None: ...
    def forward(self, x): ...

class ICLabelNet(nn.Module):
    """The ICLabel neural network."""

    img_conv: Incomplete
    psds_conv: Incomplete
    autocorr_conv: Incomplete
    conv: Incomplete
    softmax: Incomplete
    seq: Incomplete

    def __init__(self) -> None: ...
    @staticmethod
    def reshape_fortran(x: torch.Tensor, shape) -> torch.Tensor: ...
    def reshape_concat(self, tensor: torch.Tensor) -> torch.Tensor: ...
    def forward(
        self, images: torch.Tensor, psds: torch.Tensor, autocorr: torch.Tensor
    ) -> torch.Tensor: ...

def _format_input_for_torch(topo: ArrayLike, psd: ArrayLike, autocorr: ArrayLike):
    """Format the features to the correct shape and type for pytorch."""

def _run_iclabel(images: ArrayLike, psds: ArrayLike, autocorr: ArrayLike) -> NDArray:
    """Run ICLabel using onnx."""



import pytest
from mne.datasets import testing
from mne.io import read_raw
from mne.preprocessing import ICA

from tmseegpy.mne_icalabel import label_components

directory = testing.data_path() / "MEG" / "sample"
raw = read_raw(directory / "sample_audvis_trunc_raw.fif", preload=False)
raw.pick_types(eeg=True, exclude="bads")
raw.load_data()
# preprocess, fake filtering, testing dataset is filtered between [0.1, 80] Hz
with raw.info._unlock():
    raw.info["highpass"] = 1.0
    raw.info["lowpass"] = 100.0
raw.set_eeg_reference("average")


@pytest.mark.parametrize("n_components", [5, 15])
def test_label_components(n_components):
    """Simple test to check that label_components runs without raising."""
    ica = ICA(
        n_components=n_components, method="infomax", fit_params=dict(extended=True)
    )
    ica.fit(raw)
    labels = label_components(raw, ica, method="iclabel")
    assert isinstance(labels, dict)
    assert labels["y_pred_proba"].ndim == 1
    assert labels["y_pred_proba"].shape[0] == ica.n_components_


def test_label_components_with_wrong_arguments():
    """Test that wrong arguments raise."""
    ica = ICA(n_components=3, method="picard")
    ica.fit(raw)
    with pytest.raises(ValueError, match="Invalid value for the 'method' parameter"):
        label_components(raw, ica, method="101")
    with pytest.raises(ValueError, match="Invalid value for the 'method' parameter"):
        label_components(raw, ica, method="manual")



from typing import Any

class WrapStdOut:
    """Dynamically wrap to sys.stdout.

    This makes packages that monkey-patch sys.stdout (e.g.doctest, sphinx-gallery) work
    properly.
    """

    def __getattr__(self, name: str) -> Any: ...



from sklearn.preprocessing import StandardScaler
import numpy as np
import os
import tensorly as tl
from tensorly.decomposition import parafac, non_negative_parafac
from tensorly.decomposition import tucker
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, as_completed
import mne
from typing import Dict, Tuple, Optional
import warnings
from tqdm import tqdm

class TMSArtifactCleaner:
    def __init__(self, epochs: mne.Epochs, verbose: bool = True):
        """Initialize the TMS artifact cleaner."""
        self.epochs = epochs.copy()
        self.data = epochs.get_data()
        self.times = epochs.times
        self.n_epochs, self.n_channels, self.n_times = self.data.shape
        self.verbose = verbose
        self.sfreq = epochs.info['sfreq']
        
        # Initialize tensor backend
        tl.set_backend('numpy')
        
        # Normalize the data
        self.normalized_data, self.scalers = self._normalize_data(self.data)
        
        # Initialize artifact info dictionary
        self.artifact_info = {'muscle': {'times': [], 'stats': {}}}

    def _normalize_data(self, data):
        """Normalize data across channels and time points."""
        normalized_data = np.zeros_like(data)
        scalers = []
        
        # Normalize each channel separately
        for ch in range(self.n_channels):
            scaler = StandardScaler()
            ch_data = data[:, ch, :].reshape(-1, 1)
            normalized_data[:, ch, :] = scaler.fit_transform(ch_data).reshape(self.n_epochs, -1)
            scalers.append(scaler)
            
        return normalized_data, scalers

    def _detect_single_epoch(self, epoch_idx: int, 
                        start_idx: int, 
                        end_idx: int,
                        threshold_factor: float) -> Tuple[Optional[float], Optional[float]]:
        """Detect muscle artifacts in a single epoch using tensor decomposition."""
        try:
            # Extract time window of interest
            epoch_data = self.normalized_data[epoch_idx:epoch_idx+1, :, start_idx:end_idx]
            
            # Apply non-negative tensor factorization
            weights, factors = non_negative_parafac(
                epoch_data,
                rank=3,
                n_iter_max=100,
                init='random',
                random_state=42,
                tol=1e-6,
                return_errors=False
            )
            
            # The factors list contains [temporal factors, spatial factors]
            # No need to access index 2 anymore
            temporal_factors = factors[1]  # Time factors
            spatial_factors = factors[0]   # Channel factors
            
            # Compute combined artifact score
            artifact_score = np.max(np.abs(temporal_factors)) * np.max(np.abs(spatial_factors))
            
            # Detect artifact if score exceeds threshold
            if artifact_score > threshold_factor:
                return (self.times[start_idx], self.times[end_idx])
            
            return (None, None)
            
        except Exception as e:
            warnings.warn(f"Error processing epoch {epoch_idx}: {str(e)}")
            return (None, None)

    def detect_muscle_artifacts(self,
                              muscle_window: Tuple[float, float] = (0.005, 0.05),
                              threshold_factor: float = 5.0,
                              n_jobs: int = -1,
                              verbose: Optional[bool] = None) -> Dict:
        """Detect TMS-evoked muscle artifacts using parallel processing."""
        verbose = self.verbose if verbose is None else verbose
        
        # Convert time window to sample indices
        start_idx = np.searchsorted(self.times, muscle_window[0])
        end_idx = np.searchsorted(self.times, muscle_window[1])
        
        # Reset artifact info
        self.artifact_info = {'muscle': {'times': [], 'stats': {}}}
        
        # Determine number of workers
        n_jobs = min(self.n_epochs, (os.cpu_count() or 1)) if n_jobs == -1 else n_jobs
        
        # Process epochs in parallel with progress bar
        with ThreadPoolExecutor(max_workers=n_jobs) as executor:
            futures = []
            for epoch_idx in range(self.n_epochs):
                future = executor.submit(
                    self._detect_single_epoch,
                    epoch_idx,
                    start_idx,
                    end_idx,
                    threshold_factor
                )
                futures.append(future)
            
            # Collect results with progress bar
            if verbose:
                futures = tqdm(futures, total=self.n_epochs, desc="Detecting artifacts")
            
            for future in futures:
                self.artifact_info['muscle']['times'].append(future.result())
        
        # Calculate statistics
        n_detected = sum(1 for x in self.artifact_info['muscle']['times'] if x[0] is not None)
        detection_rate = n_detected / self.n_epochs
        
        self.artifact_info['muscle']['stats'].update({
            'n_detected': n_detected,
            'detection_rate': detection_rate,
            'threshold_used': threshold_factor
        })
        
        if verbose:
            print(f"\nDetected {n_detected}/{self.n_epochs} artifacts ({detection_rate*100:.1f}%)")
            
        return self.artifact_info

   
    def clean_muscle_artifacts(self,
                            n_components: int = 2,
                            verbose: Optional[bool] = None) -> mne.Epochs:
        """Clean detected muscle artifacts using tucker decomposition."""
        verbose = self.verbose if verbose is None else verbose
        cleaned_data = self.normalized_data.copy()
        
        if verbose:
            print("\nCleaning detected artifacts...")
            
        print(f"Input data shape: {self.normalized_data.shape}")
        
        for epoch_idx, (start_time, end_time) in enumerate(tqdm(
            self.artifact_info['muscle']['times'],
            disable=not verbose
        )):
            if start_time is None or end_time is None:
                continue
                
            try:
                # Get time indices
                start_idx = np.searchsorted(self.times, start_time)
                end_idx = np.searchsorted(self.times, end_time)
                
                # Extract artifact data 
                artifact_data = self.normalized_data[epoch_idx:epoch_idx+1, :, start_idx:end_idx]
                
                if artifact_data.size == 0:
                    warnings.warn(f"Empty artifact data for epoch {epoch_idx}, skipping")
                    continue
                    
                #print(f"\nEpoch {epoch_idx} artifact data shape: {artifact_data.shape}")
                
                # Apply tucker decomposition 
                ranks = [1, n_components, n_components]
                core, factors = tucker(artifact_data, rank=ranks, init='random', tol=1e-6)
                
                #print(f"Core shape: {core.shape}")
                #print(f"Factor shapes: {[f.shape for f in factors]}")
                
                # Reconstruct cleaned data
                cleaned_artifact = tl.tucker_to_tensor((core, factors))
                
                # Store cleaned data
                cleaned_data[epoch_idx:epoch_idx+1, :, start_idx:end_idx] = cleaned_artifact
                
            except Exception as e:
                warnings.warn(f"Error cleaning epoch {epoch_idx}: {str(e)}")
                continue

        # Convert back to original scale
        cleaned_data = np.array([
            scaler.inverse_transform(cleaned_data[:, i, :].reshape(-1, 1)).reshape(self.n_epochs, -1)
            for i, scaler in enumerate(self.scalers)
        ]).transpose(1, 0, 2)
        
        # Create new epochs object with cleaned data
        epochs_clean = self.epochs.copy()
        epochs_clean._data = cleaned_data
        
        # Validate results
        if self.validate_cleaning(epochs_clean):
            if verbose:
                print("\nArtifact cleaning completed successfully")
        else:
            warnings.warn("Cleaning validation failed")
            
        return epochs_clean


    def find_optimal_threshold(self,
                         muscle_window: Tuple[float, float] = (0.005, 0.05),
                         target_detection_rate: float = 0.5,
                         initial_threshold: float = 0.9,
                         max_iter: int = 100,
                         tol: float = 0.01,
                         verbose: Optional[bool] = None) -> float:
        """Find optimal threshold using binary search."""
        verbose = self.verbose if verbose is None else verbose
        
        if verbose:
            print("\nFinding optimal threshold...")
        
        # Initialize binary search
        left = 0.1  # Minimum threshold > 0
        right = initial_threshold
        best_threshold = initial_threshold
        best_cost = float('inf')
        
        for iteration in range(max_iter):
            threshold = (left + right) / 2
            
            # Test current threshold
            self.detect_muscle_artifacts(
                muscle_window=muscle_window,
                threshold_factor=threshold,
                verbose=False
            )
            detection_rate = self.artifact_info['muscle']['stats']['detection_rate']
            cost = abs(detection_rate - target_detection_rate)
            
            if verbose:
                print(f"Iteration {iteration + 1}: "
                    f"threshold={threshold:.4f}, "
                    f"detection_rate={detection_rate:.4f}, "
                    f"cost={cost:.4f}")
                
            # Update best threshold if improved
            if cost < best_cost:
                best_cost = cost
                best_threshold = threshold
                
            # Check convergence
            if cost < tol:
                break
                
            # Update search interval
            if detection_rate > target_detection_rate:
                left = threshold
            else:
                right = threshold
                
            # Stop if interval is too small
            if abs(right - left) < 1e-6:
                break
        
        if verbose:
            print(f"\nBest threshold found: {best_threshold:.4f}")
            
        return best_threshold
    
    def validate_cleaning(self, cleaned_epochs: mne.Epochs) -> bool:
        """Validate the cleaning results."""
        if cleaned_epochs.get_data().shape != self.epochs.get_data().shape:
            return False
        if np.any(np.isnan(cleaned_epochs.get_data())):
            return False
        if np.any(np.isinf(cleaned_epochs.get_data())):
            return False
            
        return True


# data_loader.py
from pathlib import Path
import mne
from typing import Optional, Union, Dict, List, Tuple
from .neurone_loader import Recording  # Keep NeurOne support

class TMSEEGLoader:
    """
    A flexible data loader for TMS-EEG data that supports multiple formats.
    Default support for NeurOne (.ses) files, with additional support for
    other common EEG formats.
    
    Parameters
    ----------
    data_path : str or Path
        Path to the data directory or file
    format : str, optional
        Data format to load ('neurone', 'brainvision', 'edf', 'cnt', 'eeglab', or 'auto')
        Default is 'neurone'
    substitute_zero_events_with : int, optional 
        Value to substitute zero events with in NeurOne data
    eeglab_montage_units : str, optional
        Units for EEGLAB channel positions ('auto', 'mm', 'm', etc.)
    """
    
    SUPPORTED_FORMATS = {
        'neurone': ('.ses',),
        'brainvision': ('.vhdr', '.eeg', '.vmrk'),
        'edf': ('.edf',),
        'cnt': ('.cnt',),
        'eeglab': ('.set',),
    }
    
    def __init__(self, 
                 data_path: Union[str, Path],
                 format: str = 'neurone',
                 substitute_zero_events_with: int = 10,
                 eeglab_montage_units: str = 'auto',
                 verbose: Optional[Union[bool, str, int]] = None):
        self.data_path = Path(data_path)
        self.format = format.lower()
        self.substitute_zero_events_with = substitute_zero_events_with
        self.eeglab_montage_units = eeglab_montage_units
        self.verbose = verbose
        
        if self.format not in self.SUPPORTED_FORMATS and self.format != 'auto':
            raise ValueError(f"Unsupported format: {format}. "
                           f"Supported formats are: {list(self.SUPPORTED_FORMATS.keys())}")
        
        # For storing loaded sessions
        self.sessions = []
        self.raw_list = []
        self.session_info = []
        
    def detect_format(self) -> str:
        """
        Detect the data format based on file extensions in the directory.
        
        Returns
        -------
        str
            Detected format name
        """
        if self.data_path.is_file():
            ext = self.data_path.suffix.lower()
            for fmt, extensions in self.SUPPORTED_FORMATS.items():
                if ext in extensions:
                    return fmt
        else:
            # Check directory contents for each format
            for fmt, extensions in self.SUPPORTED_FORMATS.items():
                if any(self.data_path.glob(f"**/*{extensions[0]}")):
                    return fmt
        
        raise ValueError("Could not detect data format")

    def load_data(self) -> List[mne.io.Raw]:
        """
        Load TMS-EEG data in the specified format.
        
        Returns
        -------
        List[mne.io.Raw]
            List of loaded Raw objects
        """
        if self.format == 'auto':
            self.format = self.detect_format()
            print(f"Detected format: {self.format}")
            
        # Dictionary mapping formats to their loading methods
        format_loaders = {
            'neurone': self._load_neurone,
            'brainvision': self._load_brainvision,
            'edf': self._load_edf,
            'cnt': self._load_cnt,
            'eeglab': self._load_eeglab
        }
        
        if self.format not in format_loaders:
            raise ValueError(f"Unsupported format: {self.format}")
            
        return format_loaders[self.format]()

    def _load_neurone(self) -> List[mne.io.Raw]:
        """
        Load NeurOne .ses files.

        Returns
        -------
        List[mne.io.Raw]
            List of loaded Raw objects
        """
        rec = Recording(str(self.data_path))
        self.sessions = rec.sessions
        self.raw_list = [
            session.to_mne(substitute_zero_events_with=self.substitute_zero_events_with)
            for session in rec.sessions
        ]
        
        # Modified session info creation
        self.session_info = []
        for session in rec.sessions:
            # Get session path as Path object and extract name
            session_path = Path(session.path)
            session_name = session_path.name if isinstance(session_path, Path) else Path(session.path).name
            
            self.session_info.append({
                'name': session_name,
                'format': 'neurone',
                'path': str(session.path)
            })
        
        return self.raw_list

    def _load_brainvision(self) -> List[mne.io.Raw]:
        """
        Load BrainVision files.
        
        Returns
        -------
        List[mne.io.Raw]
            List of loaded Raw objects
        """
        if self.data_path.is_file():
            # Single file
            raw = mne.io.read_raw_brainvision(self.data_path, preload=True, verbose=self.verbose)
            self.raw_list = [raw]
            self.session_info = [{
                'name': self.data_path.stem,
                'format': 'brainvision',
                'path': str(self.data_path)
            }]
        else:
            # Directory with multiple files
            vhdr_files = list(self.data_path.glob("**/*.vhdr"))
            self.raw_list = []
            self.session_info = []
            
            for f in vhdr_files:
                try:
                    raw = mne.io.read_raw_brainvision(f, preload=True, verbose=self.verbose)
                    self.raw_list.append(raw)
                    self.session_info.append({
                        'name': f.stem,
                        'format': 'brainvision',
                        'path': str(f)
                    })
                except Exception as e:
                    print(f"Warning: Could not load {f}: {str(e)}")
                    
        return self.raw_list

    def _load_edf(self) -> List[mne.io.Raw]:
        """
        Load EDF files.
        
        Returns
        -------
        List[mne.io.Raw]
            List of loaded Raw objects
        """
        if self.data_path.is_file():
            raw = mne.io.read_raw_edf(self.data_path, preload=True, verbose=self.verbose)
            self.raw_list = [raw]
            self.session_info = [{
                'name': self.data_path.stem,
                'format': 'edf',
                'path': str(self.data_path)
            }]
        else:
            edf_files = list(self.data_path.glob("**/*.edf"))
            self.raw_list = []
            self.session_info = []
            
            for f in edf_files:
                try:
                    raw = mne.io.read_raw_edf(f, preload=True, verbose=self.verbose)
                    self.raw_list.append(raw)
                    self.session_info.append({
                        'name': f.stem,
                        'format': 'edf',
                        'path': str(f)
                    })
                except Exception as e:
                    print(f"Warning: Could not load {f}: {str(e)}")
                    
        return self.raw_list

    def _load_cnt(self) -> List[mne.io.Raw]:
        """
        Load Neuroscan .cnt files.
        
        Returns
        -------
        List[mne.io.Raw]
            List of loaded Raw objects
        """
        if self.data_path.is_file():
            raw = mne.io.read_raw_cnt(self.data_path, preload=True, verbose=self.verbose)
            self.raw_list = [raw]
            self.session_info = [{
                'name': self.data_path.stem,
                'format': 'cnt',
                'path': str(self.data_path)
            }]
        else:
            cnt_files = list(self.data_path.glob("**/*.cnt"))
            self.raw_list = []
            self.session_info = []
            
            for f in cnt_files:
                try:
                    raw = mne.io.read_raw_cnt(f, preload=True, verbose=self.verbose)
                    self.raw_list.append(raw)
                    self.session_info.append({
                        'name': f.stem,
                        'format': 'cnt',
                        'path': str(f)
                    })
                except Exception as e:
                    print(f"Warning: Could not load {f}: {str(e)}")
                    
        return self.raw_list

    def _load_eeglab(self) -> List[mne.io.Raw]:
        """
        Load EEGLAB .set files.
        
        Returns
        -------
        List[mne.io.Raw]
            List of loaded Raw objects
        """
        if self.data_path.is_file():
            raw = mne.io.read_raw_eeglab(
                self.data_path,
                preload=True,
                montage_units=self.eeglab_montage_units,
                eog='auto',
                verbose=self.verbose
            )
            self.raw_list = [raw]
            self.session_info = [{
                'name': self.data_path.stem,
                'format': 'eeglab',
                'path': str(self.data_path)
            }]
        else:
            set_files = list(self.data_path.glob("**/*.set"))
            self.raw_list = []
            self.session_info = []
            
            for f in set_files:
                try:
                    raw = mne.io.read_raw_eeglab(
                        f,
                        preload=True,
                        montage_units=self.eeglab_montage_units,
                        eog='auto',
                        verbose=self.verbose
                    )
                    self.raw_list.append(raw)
                    self.session_info.append({
                        'name': f.stem,
                        'format': 'eeglab',
                        'path': str(f)
                    })
                except Exception as e:
                    print(f"Warning: Could not load {f}: {str(e)}")
                    
        return self.raw_list

    def get_session_names(self) -> List[str]:
        """
        Get list of session names.
        
        Returns
        -------
        List[str]
            List of session names
        """
        return [info['name'] for info in self.session_info] if self.session_info else []
    
    def get_session_info(self) -> List[Dict]:
        """
        Get detailed information about loaded sessions.
        
        Returns
        -------
        List[Dict]
            List of dictionaries containing session information
        """
        return self.session_info
    
    def print_summary(self) -> None:
        """Print a summary of loaded data."""
        print(f"\nData Format: {self.format}")
        print(f"Number of sessions: {len(self.raw_list)}")
        print("\nSession Details:")
        for i, info in enumerate(self.session_info):
            print(f"\nSession {i+1}:")
            print(f"  Name: {info['name']}")
            print(f"  Format: {info['format']}")
            print(f"  Path: {info['path']}")
            if self.raw_list[i] is not None:
                print(f"  Channels: {len(self.raw_list[i].ch_names)}")
                print(f"  Duration: {self.raw_list[i].times[-1]:.1f} seconds")
                print(f"  Sampling Rate: {self.raw_list[i].info['sfreq']} Hz")


# main.py
#!/usr/bin/env python3
import argparse
import tkinter as tk
from .gui.gui_app import TMSEEG_GUI

def main():
    parser = argparse.ArgumentParser(description='Launch TMS-EEG Analysis GUI')
    parser.add_argument('--width', type=int, default=1000,
                       help='Initial window width (default: 1000)')
    parser.add_argument('--height', type=int, default=1000,
                       help='Initial window height (default: 1000)')

    args = parser.parse_args()

    try:
        root = tk.Tk()
        root.title("TMS-EEG Analysis")
        root.geometry(f"{args.width}x{args.height}")
        app = TMSEEG_GUI(root)
        root.mainloop()
    except Exception as e:
        print(f"Error initializing GUI: {e}")
        return 1

    return 0

if __name__ == "__main__":
    main()


from typing import Any, Callable

docdict: dict[str, str]
_KEYS_MNE: tuple[str, ...]
entry: str
docdict_indented: dict[int, dict[str, str]]

def fill_doc(f: Callable[..., Any]) -> Callable[..., Any]:
    """Fill a docstring with docdict entries.

    Parameters
    ----------
    f : callable
        The function to fill the docstring of (modified in place).

    Returns
    -------
    f : callable
        The function, potentially with an updated __doc__.
    """

def _indentcount_lines(lines: list[str]) -> int:
    """Minimum indent for all lines in line list.

    >>> lines = [" one", "  two", "   three"]
    >>> indentcount_lines(lines)
    1
    >>> lines = []
    >>> indentcount_lines(lines)
    0
    >>> lines = [" one"]
    >>> indentcount_lines(lines)
    1
    >>> indentcount_lines(["    "])
    0
    """

def copy_doc(source: Callable[..., Any]) -> Callable[..., Any]:
    """Copy the docstring from another function (decorator).

    The docstring of the source function is prepepended to the docstring of the function
    wrapped by this decorator.

    This is useful when inheriting from a class and overloading a method. This decorator
    can be used to copy the docstring of the original method.

    Parameters
    ----------
    source : callable
        The function to copy the docstring from.

    Returns
    -------
    wrapper : callable
        The decorated function.

    Examples
    --------
    >>> class A:
    ...     def m1():
    ...         '''Docstring for m1'''
    ...         pass
    >>> class B(A):
    ...     @copy_doc(A.m1)
    ...     def m1():
    ...         '''this gets appended'''
    ...         pass
    >>> print(B.m1.__doc__)
    Docstring for m1 this gets appended
    """



from __future__ import annotations  # c.f. PEP 563, PEP 649

from typing import TYPE_CHECKING

from matplotlib import pyplot as plt
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg
from mne import BaseEpochs
from mne.io import BaseRaw
from mne.preprocessing import ICA
from mne.utils import _validate_type, check_version
from mne.viz import set_browser_backend
from qtpy.QtCore import Qt, Slot
from qtpy.QtWidgets import (
    QAbstractItemView,
    QButtonGroup,
    QGridLayout,
    QListWidget,
    QMainWindow,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from mne_icalabel.config import ICA_LABELS_TO_MNE

if TYPE_CHECKING:
    from typing import Union


class ICAComponentLabeler(QMainWindow):
    """Qt GUI to annotate components.

    Parameters
    ----------
    inst : Raw | Epochs
    ica : ICA
    show : bool
    """

    def __init__(
        self, inst: Union[BaseRaw, BaseEpochs], ica: ICA, show: bool = True
    ) -> None:
        ICAComponentLabeler._check_inst_ica(inst, ica)
        super().__init__()  # initialize the QMainwindow
        set_browser_backend("qt")  # force MNE to use the QT Browser

        # keep an internal pointer to the instance and to the ICA
        self._inst = inst
        self._ica = ica
        # define valid labels
        self._labels = list(ICA_LABELS_TO_MNE.keys())
        # prepare the GUI
        self._load_ui()

        # dictionary to remember selected labels, with the key as the 'indice'
        # of the component and the value as the 'label'.
        self.selected_labels: dict[int, str] = dict()

        # connect signal to slots
        self._connect_signals_to_slots()

        # select first IC
        self._selected_component = 0
        self._components_listWidget.setCurrentRow(0)  # emit signal

        if show:
            self.show()

    def _save_labels(self) -> None:
        """Save the selected labels to the ICA instance."""
        # convert the dict[int, str] to dict[str, list[int]] with the key as
        # 'label' and value as a list of component indices.
        labels2save: dict[str, list[int]] = {key: [] for key in self.labels}
        for component, label in self.selected_labels.items():
            labels2save[label].append(component)
        # sanity-check: uniqueness
        assert all(len(elt) == len(set(elt)) for elt in labels2save.values())

        for label, comp_list in labels2save.items():
            mne_label = ICA_LABELS_TO_MNE[label]
            if mne_label not in self._ica.labels_:
                self._ica.labels_[mne_label] = comp_list
                continue
            for comp in comp_list:
                if comp not in self._ica.labels_[mne_label]:
                    self._ica.labels_[mne_label].append(comp)

    # - UI --------------------------------------------------------------------
    def _load_ui(self) -> None:
        """Prepare the GUI.

        Widgets
        -------
        self._components_listWidget
        self._labels_buttonGroup
        self._mpl_widgets (dict)
            - topomap
            - psd
        self._timeSeries_widget

        Matplotlib figures
        ------------------
        self._mpl_figures (dict)
            - topomap
            - psd
        """
        self.setWindowTitle("ICA Component Labeler")
        self.setContextMenuPolicy(Qt.NoContextMenu)

        # create central widget and main layout
        self._central_widget = QWidget(self)
        self._central_widget.setObjectName("central_widget")
        grid_layout = QGridLayout()
        self._central_widget.setLayout(grid_layout)
        self.setCentralWidget(self._central_widget)

        # QListWidget with the components' names.
        self._components_listWidget = QListWidget(self._central_widget)
        self._components_listWidget.setSelectionMode(QAbstractItemView.SingleSelection)
        self._components_listWidget.addItems(
            [f"ICA{str(k).zfill(3)}" for k in range(self.n_components_)]
        )
        grid_layout.addWidget(self._components_listWidget, 0, 0, 2, 1)

        # buttons to select labels
        self._labels_buttonGroup = QButtonGroup(self._central_widget)
        buttonGroup_layout = QVBoxLayout()
        self._labels_buttonGroup.setExclusive(True)
        for k, label in enumerate(self.labels + ["Reset"]):
            pushButton = QPushButton(self._central_widget)
            pushButton.setObjectName(f"pushButton_{label.lower().replace(' ', '_')}")
            pushButton.setText(label)
            pushButton.setCheckable(True)
            pushButton.setChecked(False)
            pushButton.setEnabled(False)
            # buttons are ordered in the same order as labels
            self._labels_buttonGroup.addButton(pushButton, k)
            buttonGroup_layout.addWidget(pushButton)
        grid_layout.addLayout(buttonGroup_layout, 0, 1, 2, 1)

        # matplotlib figures
        self._mpl_figures = dict()
        self._mpl_widgets = dict()

        # topographic map
        fig, _ = plt.subplots(1, 1, figsize=(4, 4), dpi=100)
        fig.subplots_adjust(bottom=0, left=0, right=1, top=1, wspace=0, hspace=0)
        self._mpl_figures["topomap"] = fig
        self._mpl_widgets["topomap"] = FigureCanvasQTAgg(fig)
        grid_layout.addWidget(self._mpl_widgets["topomap"], 0, 2)

        # PSD
        fig, _ = plt.subplots(1, 1, figsize=(4, 4), dpi=100)
        fig.subplots_adjust(bottom=0, left=0, right=1, top=1, wspace=0, hspace=0)
        self._mpl_figures["psd"] = fig
        self._mpl_widgets["psd"] = FigureCanvasQTAgg(fig)
        grid_layout.addWidget(self._mpl_widgets["psd"], 0, 3)

        # time-series, initialized with an empty widget.
        # TODO: When the browser supports changing the instance displayed, this
        # should be initialized to a browser with the first IC.
        self._timeSeries_widget = QWidget()
        grid_layout.addWidget(self._timeSeries_widget, 1, 2, 1, 2)

    # - Checkers --------------------------------------------------------------
    @staticmethod
    def _check_inst_ica(inst: Union[BaseRaw, BaseEpochs], ica: ICA) -> None:
        """Check if the ICA was fitted."""
        _validate_type(inst, (BaseRaw, BaseEpochs), "inst", "raw or epochs")
        _validate_type(ica, ICA, "ica", "ICA")
        if ica.current_fit == "unfitted":
            raise ValueError(
                "ICA instance should be fitted on raw/epochs data before "
                "running the ICA labeling GUI. Run 'ica.fit(inst)'."
            )

    # - Properties ------------------------------------------------------------
    @property
    def inst(self) -> Union[BaseRaw, BaseEpochs]:
        """Instance on which the ICA has been fitted."""
        return self._inst

    @property
    def ica(self) -> ICA:
        """Fitted ICA decomposition."""
        return self._ica

    @property
    def n_components_(self) -> int:
        """The number of fitted components."""
        return self._ica.n_components_

    @property
    def labels(self) -> list[str]:
        """List of valid labels."""
        return self._labels

    @property
    def selected_component(self) -> int:
        """IC selected and displayed."""
        return self._selected_component

    # - Slots -----------------------------------------------------------------
    def _connect_signals_to_slots(self) -> None:
        """Connect all the signals and slots of the GUI."""
        self._components_listWidget.currentRowChanged.connect(
            self._components_listWidget_clicked
        )
        self._labels_buttonGroup.buttons()[-1].clicked.connect(self._reset)

    @Slot()
    def _components_listWidget_clicked(self) -> None:
        """Update the plots and the saved labels accordingly."""
        self._update_selected_labels()
        self._reset_buttons()

        # update selected IC
        self._selected_component = self._components_listWidget.currentRow()

        # reset matplotlib figures
        for fig in self._mpl_figures.values():
            fig.axes[0].clear()
        # create dummy figure and axes to hold the unused plots from plot_properties
        dummy_fig, dummy_axes = plt.subplots(3)
        # create axes argument provided to plot_properties
        axes = [
            self._mpl_figures["topomap"].axes[0],
            dummy_axes[0],
            dummy_axes[1],
            self._mpl_figures["psd"].axes[0],
            dummy_axes[2],
        ]
        # update matplotlib plots with plot_properties
        self.ica.plot_properties(
            self.inst, axes=axes, picks=self.selected_component, show=False
        )
        del dummy_fig
        # remove title from topomap axes
        self._mpl_figures["topomap"].axes[0].set_title("")
        # update the matplotlib canvas
        for fig in self._mpl_figures.values():
            fig.tight_layout()
            fig.canvas.draw()
            fig.canvas.flush_events()

        # swap timeSeries widget
        kwargs_plot_sources = (
            dict(splash=False) if check_version("mne", "1.6") else dict()
        )
        timeSeries_widget = self.ica.plot_sources(
            self.inst, picks=[self.selected_component], **kwargs_plot_sources
        )
        self._central_widget.layout().replaceWidget(
            self._timeSeries_widget, timeSeries_widget
        )
        self._timeSeries_widget.setParent(None)
        self._timeSeries_widget = timeSeries_widget

        # select buttons that were previously selected for this IC
        if self.selected_component in self.selected_labels:
            idx = self.labels.index(self.selected_labels[self.selected_component])
            self._labels_buttonGroup.button(idx).setChecked(True)

    def _update_selected_labels(self) -> None:
        """Update the labels saved."""
        selected = self._labels_buttonGroup.checkedButton()
        if selected is not None:
            self.selected_labels[self.selected_component] = selected.text()
        self._save_labels()  # updates the ICA instance every time

    @Slot()
    def _reset(self) -> None:
        """Remove the label from the current component."""
        self._reset_buttons()
        if self.selected_component in self.selected_labels:
            del self.selected_labels[self.selected_component]

    def _reset_buttons(self) -> None:
        """Reset all buttons."""
        self._labels_buttonGroup.setExclusive(False)
        for button in self._labels_buttonGroup.buttons():
            button.setEnabled(True)
            button.setChecked(False)
        self._labels_buttonGroup.setExclusive(True)

    def closeEvent(self, event) -> None:
        """Clean up upon closing the window.

        Update the labels since the user might have selected one for the
        currently being displayed IC.
        """
        self._update_selected_labels()
        event.accept()



from typing import Union

from mne import BaseEpochs
from mne.io import BaseRaw
from mne.preprocessing import ICA
from mne.utils import _validate_type


def _validate_inst_and_ica(inst: Union[BaseRaw, BaseEpochs], ica: ICA):
    """Make sure that the provided instance and ICA are valid."""
    _validate_type(inst, (BaseRaw, BaseEpochs), "inst", "Raw or Epochs")
    _validate_ica(ica)


def _validate_ica(ica: ICA):
    """Make sure that the provided ICA is valid."""
    _validate_type(ica, ICA, "ica")
    if ica.current_fit == "unfitted":
        raise RuntimeError(
            "The provided ICA instance was not fitted. Please use the '.fit()' method "
            "to determine the independent components before trying to label them."
        )



from __future__ import annotations  # c.f. PEP 563, PEP 649

from importlib import import_module
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from typing import Callable


def requires_module(name: str):  # pragma: no cover
    """Skip a test if package is not available (decorator)."""
    try:
        import_module(name)
        skip = False
    except ImportError:
        skip = True

    def decorator(function: Callable):
        return pytest.mark.skipif(
            skip, reason=f"Test {function.__name__} skipped, requires {name}."
        )(function)

    return decorator



from mne import BaseEpochs as BaseEpochs
from mne.io import BaseRaw as BaseRaw
from mne.preprocessing import ICA

from .features import get_iclabel_features as get_iclabel_features
from .network import run_iclabel as run_iclabel

def iclabel_label_components(
    inst: BaseRaw | BaseEpochs,
    ica: ICA,
    inplace: bool = True,
    backend: str | None = None,
):
    """Label the provided ICA components with the ICLabel neural network.

    ICLabel is designed to classify ICs fitted with an extended infomax ICA
    decomposition algorithm on EEG datasets referenced to a common average and
    filtered between [1., 100.] Hz. It is possible to run ICLabel on datasets that
    do not meet those specification, but the classification performance
    might be negatively impacted. Moreover, the ICLabel paper did not study the
    effects of these preprocessing steps.

    ICLabel uses 3 features:

    - Topographic maps, based on the ICA decomposition.
    - Power Spectral Density (PSD), based on the ICA decomposition and the
      provided instance.
    - Autocorrelation, based on the ICA decomposition and the provided
      instance.

    For more information, see :footcite:t:`PionTonachini2019`.

    Parameters
    ----------
    inst : Raw | Epochs
        Instance used to fit the ICA decomposition. The instance should be
        referenced to a common average and bandpass filtered between 1 and
        100 Hz.
    ica : ICA
        ICA decomposition of the provided instance. The ICA decomposition
        should use the extended infomax method.
    inplace : bool
        Whether to modify the ``ica`` instance in place by adding the automatic
        annotations to the ``labels_`` property. By default True.
    backend : None | ``torch`` | ``onnx``
        Backend to use to run ICLabel. If None, returns the first available backend in
        the order ``torch``, ``onnx``.

    Returns
    -------
    labels_pred_proba : numpy.ndarray of shape (n_components, n_classes)
        The estimated corresponding predicted probabilities of output classes
        for each independent component. Columns are ordered with 'brain',
        'muscle artifact', 'eye blink', 'heart beat', 'line noise',
        'channel noise', 'other'.

    References
    ----------
    .. footbibliography::
    """



from __future__ import annotations  # c.f. PEP 563, PEP 649

from typing import TYPE_CHECKING

import numpy as np
from mne.utils import _validate_type
from mne.utils.check import _check_option

from .config import ICALABEL_METHODS
from .iclabel._config import ICLABEL_NUMERICAL_TO_STRING
from .utils._checks import _validate_inst_and_ica

if TYPE_CHECKING:
    from typing import Union

    from mne import BaseEpochs
    from mne.io import BaseRaw
    from mne.preprocessing import ICA


def label_components(inst: Union[BaseRaw, BaseEpochs], ica: ICA, method: str):
    """Automatically label the ICA components with the selected method.

    Parameters
    ----------
    inst : Raw | Epochs
        The data instance used to fit the ICA instance.
    ica : ICA
        The fitted ICA instance.
    method : str
        The proposed method for labeling components. Must be one of:
        ``'iclabel'``.

    Returns
    -------
    component_dict : dict
        A dictionary with the following fields:

        - 'y_pred_proba' : array of shape (n_components,)
              Estimated predicted probability of the output class
              for each independent component.
        - 'labels': list of shape (n_components,)
              The corresponding string label of each class in 'y_pred'.

    Notes
    -----
    Please refer to the following function for additional information on each
    method:

    - ``'iclabel'``: `~mne_icalabel.iclabel.iclabel_label_components`
    """
    _validate_type(method, str, "method")
    _check_option(
        "method", method, [elt for elt in ICALABEL_METHODS if elt != "manual"]
    )

    _validate_inst_and_ica(inst, ica)
    labels_pred_proba = ICALABEL_METHODS[method](inst, ica)  # type: ignore
    labels_pred = np.argmax(labels_pred_proba, axis=1)
    labels = [ICLABEL_NUMERICAL_TO_STRING[label] for label in labels_pred]
    assert ica.n_components_ == labels_pred.size  # sanity-check
    assert ica.n_components_ == labels_pred_proba.shape[0]  # sanity-check
    y_pred_proba = labels_pred_proba[np.arange(ica.n_components_), labels_pred]

    component_dict = {
        "y_pred_proba": y_pred_proba,
        "labels": labels,
    }
    return component_dict



# setup.py
from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="tmseegpy",
    version="0.1.0",
    author="LazyCyborg",
    author_email="hjarneko@gmail.com",
    description="A pipeline for preprocessing and analyzing TMS-EEG data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/LazyCyborg/tmseegpy",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
    install_requires=[
        'numpy >= 1.16.0',
        'scipy >= 1.2.0',
        'mne >= 1.1',
        'pooch',
        'torch',
        "scikit-learn",
        "tensorly",
        "matplotlib",
        "seaborn",
        "tqdm",
        "mne-faster",
        'psutil',
        'construct',
        'importlib-metadata; python_version<"3.8"',
        'importlib-resources; python_version<"3.9"',

    ],
    entry_points={
        "console_scripts": [
            "tmseegpy=tmseegpy.gui:main",
        ],
    },
)


# tmseegpy/__init__.py
from .gui.gui_app import TMSEEG_GUI
from .analyze import *
from .clean import *
from .pcist import *
from .preproc import *
from .preproc_vis import *
from .run import *
from .validate_tep import *
from .dataloader import *

__version__ = "0.1.0"



from mne import BaseEpochs
from mne.io import BaseRaw
from mne.preprocessing import ICA

def _validate_inst_and_ica(inst: BaseRaw | BaseEpochs, ica: ICA):
    """Make sure that the provided instance and ICA are valid."""

def _validate_ica(ica: ICA):
    """Make sure that the provided ICA is valid."""



# microstates.py
# Might be fun to check microstate transitions before and after TMS? Will probably need many measurements to get to some stable microstates? 
import numpy as np
from typing import Optional, Union, Dict, List, Tuple
import matplotlib.pyplot as plt
import mne
from pycrostates.preprocessing import extract_gfp_peaks, resample
from pycrostates.cluster import ModKMeans
from pycrostates.segmentation import EpochsSegmentation
from pycrostates.io import ChData
from pycrostates.viz import plot_cluster_centers

class Microstate:
    """
    Class implementation of Microstate Analysis for EEG epochs using pycrostates.
    Supports combined analysis across multiple recordings.
    """
    
    def __init__(self):
        self.epochs_list = []  # Store multiple recordings
        self.segmentations = {}  # Store segmentations for each recording
        self.global_clustering = None  # Store global clustering results
        
    def add_recording(self, epochs: mne.epochs.Epochs, recording_id: str):
        """Add a recording to the analysis."""
        if not isinstance(epochs, mne.epochs.Epochs):
            raise ValueError("Input must be an MNE Epochs object")
        self.epochs_list.append((recording_id, epochs))
        
    def perform_global_clustering(self, 
                                n_clusters: int = 5,
                                n_resamples: int = 3,
                                n_samples: int = 1000,
                                picks: str = 'eeg',
                                min_peak_distance: int = 1,
                                random_state: int = 42,
                                n_jobs: int = -1,
                                return_details: bool = True) -> Union[Dict, float]:
        """
        Performs microstate analysis across all recordings with GFP peak extraction and resampling.
        """
        if not self.epochs_list:
            raise ValueError("No recordings added. Use add_recording() first.")
            
        # Initialize storage for all recordings
        all_gfp_data = []
        all_resamples = []
        all_info = None
        
        # Process each recording
        for recording_id, epochs in self.epochs_list:
            # Extract GFP peaks
            gfp_data = extract_gfp_peaks(
                epochs,
                picks=picks,
                min_peak_distance=min_peak_distance,
                return_all=False
            )
            
            # Store channel info from first recording
            if all_info is None:
                all_info = gfp_data.info
                
            # Perform resampling
            recording_resamples = resample(
                gfp_data,
                n_resamples=n_resamples,
                n_samples=n_samples,
                random_state=random_state
            )
            
            all_gfp_data.append(gfp_data)
            all_resamples.extend(recording_resamples)
        
        # Initialize storage for clustering results
        resample_results = []
        gev_scores = []
        
        # Perform clustering on each resample
        for resamp in all_resamples:
            mod_kmeans = ModKMeans(
                n_clusters=n_clusters,
                random_state=random_state
            )
            mod_kmeans.fit(resamp, n_jobs=n_jobs, verbose=True)
            
            resample_results.append(mod_kmeans.cluster_centers_)
            gev_scores.append(mod_kmeans.GEV_)
        
        # Compute final clustering on concatenated centers
        all_resampling_results = np.vstack(resample_results).T
        all_resampling_results = ChData(all_resampling_results, all_info)
        
        final_clustering = ModKMeans(
            n_clusters=n_clusters,
            random_state=random_state
        )
        final_clustering.fit(
            all_resampling_results,
            n_jobs=n_jobs,
            verbose="WARNING"
        )
        
        # Store global clustering results
        self.global_clustering = final_clustering
        
        # Backfit to individual recordings
        for recording_id, epochs in self.epochs_list:
            self.segmentations[recording_id] = final_clustering.predict(
                epochs,
                picks=picks
            )
        
        if return_details:
            details = {
                'resample_results': resample_results,
                'gev_scores': gev_scores,
                'mean_gev': np.mean(gev_scores),
                'std_gev': np.std(gev_scores),
                'final_clusters': final_clustering.cluster_centers_,
                'final_gev': final_clustering.GEV_,
                'all_gfp_data': all_gfp_data,
                'resamples': all_resamples,
                'segmentations': self.segmentations,
                'picks': picks
            }
            return details
        
        return np.mean(gev_scores)
    
    def get_recording_segmentation(self, recording_id: str) -> Optional[EpochsSegmentation]:
        """Get segmentation for a specific recording."""
        return self.segmentations.get(recording_id)
    
    def compute_recording_parameters(self, 
                                   recording_id: str,
                                   norm_gfp: bool = True, 
                                   return_dist: bool = False) -> Dict:
        """Compute microstate parameters for a specific recording."""
        segmentation = self.get_recording_segmentation(recording_id)
        if segmentation is None:
            raise ValueError(f"No segmentation found for recording {recording_id}")
        return segmentation.compute_parameters(norm_gfp=norm_gfp, return_dist=return_dist)
    
    def compute_recording_transitions(self,
                                    recording_id: str,
                                    stat: str = 'probability',
                                    ignore_repetitions: bool = True) -> Tuple[np.ndarray, np.ndarray]:
        """Compute transition matrices for a specific recording."""
        segmentation = self.get_recording_segmentation(recording_id)
        if segmentation is None:
            raise ValueError(f"No segmentation found for recording {recording_id}")
            
        observed = segmentation.compute_transition_matrix(
            stat=stat,
            ignore_repetitions=ignore_repetitions
        )
        
        expected = segmentation.compute_expected_transition_matrix(
            stat=stat,
            ignore_repetitions=ignore_repetitions
        )
        
        return observed, expected
    
    def plot_recording_segmentation(self, 
                                  recording_id: str,
                                  cmap=None, 
                                  axes=None, 
                                  block=False, 
                                  show=None) -> plt.Figure:
        """Plot segmentation for a specific recording."""
        segmentation = self.get_recording_segmentation(recording_id)
        if segmentation is None:
            raise ValueError(f"No segmentation found for recording {recording_id}")
            
        if axes is None:
            fig, axes = plt.subplots(figsize=(12, 6))
        else:
            fig = axes.figure
            
        labels = segmentation.labels
        
        if cmap is None:
            colors = ['navy', 'darkturquoise', 'forestgreen', 'gold', 'darkorange']
            cmap = plt.cm.colors.ListedColormap(colors[:labels.max()+1])
        
        im = axes.imshow(labels.T, aspect='auto', cmap=cmap, interpolation='nearest')
        
        axes.set_title(f'Microstate Segmentation - Recording {recording_id}', fontsize=12, pad=10)
        axes.set_xlabel('Epochs', fontsize=10)
        axes.set_ylabel('Time Points', fontsize=10)
        
        axes.set_xticks([])
        axes.grid(False)
        
        cbar = plt.colorbar(im, ax=axes)
        cbar.set_label('Microstate', fontsize=10)
        
        n_states = labels.max() + 1
        cbar.set_ticks(np.arange(n_states))
        cbar.set_ticklabels([f'State {i}' for i in range(n_states)])
        
        text = (
            f"Recording: {recording_id}\n"
            "Colors indicate different brain states\n"
            f"Analysis identified {n_states} distinct states"
        )
        plt.figtext(0.02, 0.02, text, fontsize=8, style='italic',
                bbox=dict(facecolor='white', edgecolor='gray', alpha=0.8, pad=5))
        
        plt.tight_layout()
        
        if show:
            plt.show()
            
        return fig
    
    def plot_final_clusters(self, details: Dict) -> plt.Figure:
        """Plot the final clustering results."""
        if self.global_clustering is None:
            raise ValueError("Must run perform_global_clustering first")
            
        # Create figure and axes
        n_clusters = details['final_clusters'].shape[0]
        f, axes = plt.subplots(1, n_clusters, figsize=(4*n_clusters, 3))
        
        # Use info from first recording's GFP data
        info = details['all_gfp_data'][0].info
        
        # Plot cluster centers
        return plot_cluster_centers(
            cluster_centers=details['final_clusters'],
            info=info,
            axes=axes
        )
    
    def compare_pre_post_tms(self,
                            recording_id: str,
                            pre_window: Tuple[float, float] = (-500, -10),
                            post_window: Tuple[float, float] = (10, 500),
                            ignore_repetitions: bool = False,
                            log_base: Union[float, str] = 2,
                            return_details: bool = True) -> Union[Dict, float]:
        """Compare microstate characteristics before and after TMS pulse for a specific recording."""
        segmentation = self.get_recording_segmentation(recording_id)
        if segmentation is None:
            raise ValueError(f"No segmentation found for recording {recording_id}")
            
        # Get recording data
        _, epochs = next((rec for rec in self.epochs_list if rec[0] == recording_id), (None, None))
        if epochs is None:
            raise ValueError(f"Recording {recording_id} not found")
            
        # Get original times and labels
        orig_times = epochs.times * 1000
        orig_labels = segmentation.labels
        
        # Find indices for pre and post windows
        pre_start_idx = np.searchsorted(orig_times, pre_window[0])
        pre_end_idx = np.searchsorted(orig_times, pre_window[1])
        post_start_idx = np.searchsorted(orig_times, post_window[0])
        post_end_idx = np.searchsorted(orig_times, post_window[1])
        
        # Extract times and labels for pre and post periods
        pre_times = orig_times[pre_start_idx:pre_end_idx]
        post_times = orig_times[post_start_idx:post_end_idx]
        
        pre_labels = orig_labels[:, pre_start_idx:pre_end_idx]
        post_labels = orig_labels[:, post_start_idx:post_end_idx]
        
        # Calculate entropy directly from labels
        def calc_entropy(labels, ignore_reps=False, base=2):
            # Flatten labels across epochs
            flat_labels = labels.reshape(-1)
            if ignore_reps:
                # Remove consecutive repeats
                mask = np.diff(flat_labels, prepend=flat_labels[0]-1) != 0
                flat_labels = flat_labels[mask]
            
            # Calculate probabilities
            unique, counts = np.unique(flat_labels, return_counts=True)
            probs = counts / len(flat_labels)
            
            # Calculate entropy
            return -np.sum(probs * np.log(probs) / np.log(base))
        
        pre_entropy = calc_entropy(pre_labels, ignore_reps=ignore_repetitions, base=log_base)
        post_entropy = calc_entropy(post_labels, ignore_reps=ignore_repetitions, base=log_base)
        
        # Calculate transition matrices
        def calc_transition_matrix(labels, ignore_reps=False):
            n_states = len(np.unique(labels))
            trans_mat = np.zeros((n_states, n_states))
            
            # Process each epoch separately
            for epoch_labels in labels:
                if ignore_reps:
                    mask = np.diff(epoch_labels, prepend=epoch_labels[0]-1) != 0
                    epoch_labels = epoch_labels[mask]
                
                # Count transitions
                for i in range(len(epoch_labels)-1):
                    from_state = epoch_labels[i]
                    to_state = epoch_labels[i+1]
                    trans_mat[from_state, to_state] += 1
            
            # Convert to probabilities
            row_sums = trans_mat.sum(axis=1, keepdims=True)
            row_sums[row_sums == 0] = 1  # Avoid division by zero
            trans_mat = trans_mat / row_sums
            
            return trans_mat
        
        pre_trans = calc_transition_matrix(pre_labels, ignore_reps=ignore_repetitions)
        post_trans = calc_transition_matrix(post_labels, ignore_reps=ignore_repetitions)
        
        # Calculate basic statistics for each microstate
        def calc_state_stats(labels):
            n_states = len(np.unique(labels.reshape(-1)))
            stats = {}
            
            # Process each state
            for state in range(n_states):
                state_name = f"state_{state}"
                
                # Calculate overall occurrence rate
                mask = (labels == state)
                stats[f"{state_name}_occurrence"] = np.mean(mask)
                
                # Calculate mean duration for each epoch, then average
                durations = []
                for epoch_labels in labels:
                    epoch_mask = (epoch_labels == state)
                    if np.any(epoch_mask):
                        # Find runs of True values
                        runs = np.diff(np.concatenate(([False], epoch_mask, [False])).astype(int))
                        starts = np.where(runs == 1)[0]
                        ends = np.where(runs == -1)[0]
                        durations.extend(ends - starts)
                
                if durations:
                    stats[f"{state_name}_mean_duration"] = np.mean(durations)
                else:
                    stats[f"{state_name}_mean_duration"] = 0
                    
            return stats
        
        pre_stats = calc_state_stats(pre_labels)
        post_stats = calc_state_stats(post_labels)
        
        if return_details:
            details = {
                'pre_window': pre_window,
                'post_window': post_window,
                'pre_entropy': pre_entropy,
                'post_entropy': post_entropy,
                'entropy_difference': post_entropy - pre_entropy,
                'pre_transition_matrix': pre_trans,
                'post_transition_matrix': post_trans,
                'pre_statistics': pre_stats,
                'post_statistics': post_stats,
                'pre_labels': pre_labels,
                'post_labels': post_labels,
                'pre_times': pre_times,
                'post_times': post_times
            }
            return details
        
        return post_entropy - pre_entropy

    def plot_pre_post_comparison(self, recording_id: str, comparison_details: Dict) -> plt.Figure:
        """Plot comparison of pre and post TMS microstate characteristics for a specific recording."""
        # Create figure with adjusted size
        fig = plt.figure(figsize=(12, 8))
        gs = plt.GridSpec(2, 2, height_ratios=[1, 1.2], hspace=0.4, wspace=0.3)
        
        # Update title to include recording ID
        title_suffix = f' - Recording {recording_id}'

            
        # Plot entropy comparison
        ax1 = fig.add_subplot(gs[0, 0])
        entropy_bars = ax1.bar(['Pre-TMS', 'Post-TMS'], 
                            [comparison_details['pre_entropy'], 
                            comparison_details['post_entropy']],
                            color=['lightblue', 'darkblue'],
                            width=0.5)
        ax1.set_title('Microstate Entropy\n(Higher = More Complex/Random)', fontsize=10, pad=10)
        ax1.set_ylabel('Shannon Entropy (bits)')
        
        # Add value labels on bars with increased spacing
        for bar in entropy_bars:
            height = bar.get_height()
            ax1.text(bar.get_x() + bar.get_width()/2., height + 0.05,
                    f'{height:.3f}',
                    ha='center', va='bottom')
        
        # Add entropy change text
        change = comparison_details['entropy_difference']
        change_text = f"Change: {change:+.3f} bits"
        ax1.text(0.5, -0.2, change_text, ha='center', transform=ax1.transAxes,
                fontsize=9, style='italic')
        
        # Plot transition matrix
        ax2 = fig.add_subplot(gs[0, 1])
        pre_trans = comparison_details['pre_transition_matrix']
        im_pre = ax2.imshow(pre_trans, cmap='Blues', aspect='auto')
        ax2.set_title('Pre-TMS State Transitions\n(Darker = More Frequent)', fontsize=10, pad=10)
        cbar = plt.colorbar(im_pre, ax=ax2)
        cbar.set_label('Transition Probability', fontsize=8)
        ax2.set_xlabel('To State')
        ax2.set_ylabel('From State')
        
        # Statistics plot with simplified design
        ax3 = fig.add_subplot(gs[1, :])
        
        # Extract statistics
        pre_stats = comparison_details['pre_statistics']
        post_stats = comparison_details['post_statistics']
        
        # Get all unique state numbers
        all_states = sorted(list(set(int(key.split('_')[1]) for key in pre_stats.keys())))
        n_states = len(all_states)
        
        # Create grouped bar plot
        x = np.arange(n_states)
        width = 0.35
        
        # Plot occurrence rates
        pre_occurrence = [pre_stats.get(f'state_{state}_occurrence', 0) for state in all_states]
        post_occurrence = [post_stats.get(f'state_{state}_occurrence', 0) for state in all_states]
        
        bars1 = ax3.bar(x - width/2, pre_occurrence, width, label='Pre-TMS',
                    color='lightblue', alpha=0.7)
        bars2 = ax3.bar(x + width/2, post_occurrence, width, label='Post-TMS',
                    color='darkblue', alpha=0.7)
        
        # Customize the plot
        ax3.set_ylabel('Occurrence Rate\n(proportion of time)', fontsize=9)
        ax3.set_title('Microstate Occurrence Rates\n(How often each state appears)', fontsize=10, pad=10)
        ax3.set_xticks(x)
        ax3.set_xticklabels([f'State {i}' for i in all_states])
        ax3.legend(loc='upper right')
        ax3.grid(True, axis='y', linestyle='--', alpha=0.3)
        
        # Add value labels with reduced fontsize
        def autolabel(bars):
            for bar in bars:
                height = bar.get_height()
                ax3.text(bar.get_x() + bar.get_width()/2., height,
                        f'{height:.3f}',
                        ha='center', va='bottom', fontsize=8)
        
        autolabel(bars1)
        autolabel(bars2)
        
        # Add detailed explanatory text
        explanation = (
            "Microstate Analysis Results:\n"
            "• Entropy: Measures the randomness/complexity of state sequences\n"
            "• Transitions: Shows how likely states change from one to another\n"
            "• Occurrence Rate: Shows the proportion of time each state is active\n"
            f"Analysis identified {n_states} distinct brain states"
        )
        
        fig.text(0.5, 0.02, explanation, ha='center', va='center',
                fontsize=8, style='italic',
                bbox=dict(facecolor='white', edgecolor='gray', alpha=0.8, pad=5))
        
        plt.tight_layout()
        # Adjust layout to make room for text
        plt.subplots_adjust(bottom=0.2)
        return fig


import pytest
from mne.datasets import testing
from mne.io import BaseRaw, read_raw
from mne.preprocessing import ICA

from tmseegpy.mne_icalabel.gui import label_ica_components

directory = testing.data_path() / "MEG" / "sample"
raw = read_raw(directory / "sample_audvis_trunc_raw.fif", preload=False)
raw.pick_types(eeg=True, exclude="bads")
raw.load_data()
# preprocess, fake filtering, testing dataset is filtered between [0.1, 80] Hz
with raw.info._unlock():
    raw.info["highpass"] = 1.0
    raw.info["lowpass"] = 100.0
ica = ICA(n_components=5, random_state=12345, fit_params=dict(tol=1e-1))
ica.fit(raw)


def test_label_components_gui_display():
    """Test that the GUI is displayed correctly."""
    ica_ = ica.copy()
    gui = label_ica_components(raw, ica_, show=False)
    # test setting the label
    assert isinstance(gui.inst, BaseRaw)
    assert isinstance(gui.ica, ICA)
    assert gui.n_components_ == ica.n_components_
    # the initial component should be 0
    assert gui.selected_component == 0


def test_invalid_arguments():
    """Test error error raised with invalid arguments."""
    with pytest.raises(TypeError, match="ica must be an instance of ICA"):
        label_ica_components(raw, 101)

    with pytest.raises(TypeError, match="inst must be an instance of raw or epochs"):
        label_ica_components(101, ica)

    with pytest.raises(ValueError, match="ICA instance should be fitted"):
        label_ica_components(raw, ICA(n_components=10, random_state=12345))



from __future__ import annotations  # c.f. PEP 563, PEP 649

from typing import TYPE_CHECKING

from mne.utils import _check_option

from ...utils._imports import import_optional_dependency

if TYPE_CHECKING:
    from typing import Optional

    from numpy.typing import ArrayLike, NDArray


def run_iclabel(
    images: ArrayLike,
    psds: ArrayLike,
    autocorr: ArrayLike,
    backend: Optional[str] = "pytorch",
) -> NDArray:
    """Run the ICLabel network on the provided set of features.

    The features are un-formatted and are as-returned by
    `~mne_icalabel.iclabel.get_iclabel_features`. For more information,
    see :footcite:t:`PionTonachini2019`.

    Parameters
    ----------
    images : array of shape (n_components, 1, 32, 32)
        The topoplot images.
    psds : array of shape (n_components, 1, 1, 100)
        The power spectral density features.
    autocorr : array of shape (n_components, 1, 1, 100)
        The autocorrelation features.
    backend : None | ``torch`` | ``onnx``
        Backend to use to run ICLabel. If None, returns the first available backend in
        the order ``torch``, ``onnx``.

    Returns
    -------
    labels : array of shape (n_components, n_classes)
        The predicted numerical probability values for all labels in ICLabel output.
        Columns are ordered with ``'Brain'``, ``'Muscle'``, ``'Eye'``,
        ``'Heart'``, ``'Line Noise'``, ``'Channel Noise'``, and ``'Other'``.

    References
    ----------
    .. footbibliography::
    """
    _check_option("backend", backend, (None, "torch", "onnx"))
    if backend is None:
        torch = import_optional_dependency("torch", raise_error=False)
        onnx = import_optional_dependency("onnxruntime", raise_error=False)

        if torch is not None:
            from .torch import _run_iclabel
        elif torch is None and onnx is not None:
            from .onnx import _run_iclabel  # type: ignore
        else:
            raise ImportError(
                "Missing optional dependency. ICLabel requires either pytorch or "
                "onnxruntime. Use pip or conda to install one of them."
            )
    elif backend == "torch":
        import_optional_dependency("torch", raise_error=True)

        from .torch import _run_iclabel  # type: ignore
    elif backend == "onnx":
        import_optional_dependency("onnxruntime", raise_error=True)

        from .onnx import _run_iclabel  # type: ignore
    return _run_iclabel(images, psds, autocorr)



import warnings
from pathlib import Path

import numpy as np
import pytest
from mne import read_epochs_eeglab
from mne.io import read_raw
from mne.io.eeglab.eeglab import _check_load_mat
from mne.preprocessing import read_ica_eeglab
from scipy.io import loadmat

from tmseegpy.mne_icalabel.datasets import icalabel
from tmseegpy.mne_icalabel.iclabel._utils import _mne_to_eeglab_locs
from tmseegpy.mne_icalabel.iclabel.features import (
    _compute_ica_activations,
    _eeg_autocorr,
    _eeg_autocorr_fftw,
    _eeg_autocorr_welch,
    _eeg_rpsd_compute_psdmed,
    _eeg_rpsd_constants,
    _eeg_rpsd_format,
    _eeg_topoplot,
    _resample,
    _retrieve_eeglab_icawinv,
    _topoplotFast,
    get_iclabel_features,
)

dataset_path = icalabel.data_path() / "iclabel"

# Raw/Epochs files with ICA decomposition
raw_eeglab_path = dataset_path / "datasets/sample-raw.set"
raw_short_eeglab_path = dataset_path / "datasets/sample-short-raw.set"
raw_very_short_eeglab_path = dataset_path / "datasets/sample-very-short-raw.set"
epo_eeglab_path = dataset_path / "datasets/sample-epo.set"
epo_long_eeglab_path = dataset_path / "datasets/sample-long-epo.set"

# ICA activation matrix for raw/epochs
raw_icaact_eeglab_path = dataset_path / "icaact/icaact-raw.mat"
epo_icaact_eeglab_path = dataset_path / "icaact/icaact-epo.mat"

# Topography
raw_topo1_path = dataset_path / "topo/topo1-raw.mat"
epo_topo1_path = dataset_path / "topo/topo1-epo.mat"
raw_topo_feature_path = dataset_path / "topo/topo-feature-raw.mat"
epo_topo_feature_path = dataset_path / "topo/topo-feature-epo.mat"

# PSD
psd_constants_raw_path = dataset_path / "psd/constants-raw.mat"
psd_steps_raw_path = dataset_path / "psd/psd-step-by-step-raw.mat"
psd_raw_path = dataset_path / "psd/psd-raw.mat"
psd_constants_epo_path = dataset_path / "psd/constants-epo.mat"
psd_steps_epo_path = dataset_path / "psd/psd-step-by-step-epo.mat"
psd_epo_path = dataset_path / "psd/psd-epo.mat"
psd_constants_long_epo_path = dataset_path / "psd/constants-long-epo.mat"
psd_steps_long_epo_path = dataset_path / "psd/psd-step-by-step-long-epo.mat"
psd_long_epo_path = dataset_path / "psd/psd-long-epo.mat"

# Autocorrelations
autocorr_raw_path = dataset_path / "autocorr/autocorr-raw.mat"
autocorr_short_raw_path = dataset_path / "autocorr/autocorr-short-raw.mat"
autocorr_very_short_raw_path = dataset_path / "autocorr/autocorr-very-short-raw.mat"
autocorr_epo_path = dataset_path / "autocorr/autocorr-epo.mat"

# Complete features
features_raw_path = dataset_path / "features/features-raw.mat"
features_epo_path = dataset_path / "features/features-epo.mat"

# General readers
reader = {"raw": read_raw, "epo": read_epochs_eeglab}
kwargs = {"raw": dict(preload=True), "epo": dict()}


# ----------------------------------------------------------------------------
@pytest.mark.filterwarnings("ignore:Estimated head radius.*:RuntimeWarning")
@pytest.mark.parametrize(
    ("file", "psd_constant_file", "eeglab_feature_file"),
    [
        (raw_eeglab_path, psd_constants_raw_path, features_raw_path),
        (epo_eeglab_path, psd_constants_epo_path, features_epo_path),
    ],
)
def test_get_features_from_precomputed_ica(
    file, psd_constant_file, eeglab_feature_file
):
    """Test that we get the correct set of features from an MNE instance.

    Corresponds to the output from 'ICL_feature_extractor.m'.
    """
    type_ = str(Path(file).stem)[-3:]
    inst = reader[type_](file, **kwargs[type_])
    ica = read_ica_eeglab(file)

    # Retrieve topo and autocorr
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        topo, _, autocorr = get_iclabel_features(inst, ica)

    # Build PSD feature manually to match the subset
    # retrieve activation
    icaact = _compute_ica_activations(inst, ica)
    # retrieve subset from eeglab
    constants_eeglab = loadmat(psd_constant_file)["constants"][0, 0]
    assert constants_eeglab["subset"].shape[0] == 1
    subset_eeglab = constants_eeglab["subset"][0, :] - 1
    # retrieve constants from python
    ncomp, nfreqs, n_points, nyquist, index, window, _ = _eeg_rpsd_constants(inst, ica)
    # compute psd
    psdmed = _eeg_rpsd_compute_psdmed(
        inst, icaact, ncomp, nfreqs, n_points, nyquist, index, window, subset_eeglab
    )
    psd = _eeg_rpsd_format(psdmed)
    psd *= 0.99

    # Compare with MATLAB
    features_eeglab = loadmat(eeglab_feature_file)["features"]
    topo_eeglab = features_eeglab[0, 0]
    psd_eeglab = features_eeglab[0, 1]
    autocorr_eeglab = features_eeglab[0, 2]
    assert np.allclose(topo, topo_eeglab)
    assert np.allclose(psd, psd_eeglab, atol=1e-6)
    # the autocorr for epoch does not have the same tolerance
    atol = 1e-8 if type_ == "raw" else 1e-4
    assert np.allclose(autocorr, autocorr_eeglab, atol=atol)


# ----------------------------------------------------------------------------
@pytest.mark.filterwarnings("ignore:Estimated head radius.*:RuntimeWarning")
@pytest.mark.parametrize("file", [raw_eeglab_path, epo_eeglab_path])
def test_retrieve_eeglab_icawinv(file):
    """Test that the icawinv is correctly retrieved from an MNE ICA object."""
    ica = read_ica_eeglab(file)
    icawinv, _ = _retrieve_eeglab_icawinv(ica)

    eeg = _check_load_mat(file, None)
    assert np.allclose(icawinv, eeg.icawinv)


@pytest.mark.filterwarnings("ignore:Estimated head radius.*:RuntimeWarning")
@pytest.mark.parametrize(
    ("file", "eeglab_result_file"),
    [
        (raw_eeglab_path, raw_icaact_eeglab_path),
        (epo_eeglab_path, epo_icaact_eeglab_path),
    ],
)
def test_compute_ica_activations(file, eeglab_result_file):
    """Test that the icaact is correctly retrieved from an MNE ICA object."""
    type_ = str(Path(file).stem)[-3:]
    inst = reader[type_](file, **kwargs[type_])
    ica = read_ica_eeglab(file)
    icaact = _compute_ica_activations(inst, ica)

    icaact_eeglab = loadmat(eeglab_result_file)["icaact"]
    assert np.allclose(icaact, icaact_eeglab, atol=1e-4)


# ----------------------------------------------------------------------------
@pytest.mark.filterwarnings("ignore:Estimated head radius.*:RuntimeWarning")
@pytest.mark.parametrize(
    ("file", "eeglab_result_file"),
    [(raw_eeglab_path, raw_topo1_path), (epo_eeglab_path, epo_topo1_path)],
)
def test_topoplotFast(file, eeglab_result_file):
    """Test topoplotFast on a single component."""
    # load inst
    type_ = str(Path(file).stem)[-3:]
    inst = reader[type_](file, **kwargs[type_])
    ica = read_ica_eeglab(file)
    # convert coordinates
    rd, th = _mne_to_eeglab_locs(inst, ica.ch_names)
    th = np.pi / 180 * th
    # get icawinv
    icawinv, _ = _retrieve_eeglab_icawinv(ica)
    # compute topo feature for the first component
    topo1 = _topoplotFast(icawinv[:, 0], rd, th)
    # load from eeglab
    topo1_eeglab = loadmat(eeglab_result_file)["topo1"]
    # convert nan to num
    assert np.allclose(topo1, topo1_eeglab, equal_nan=True)


@pytest.mark.filterwarnings("ignore:Estimated head radius.*:RuntimeWarning")
@pytest.mark.parametrize(
    ("file", "eeglab_result_file"),
    [
        (raw_eeglab_path, raw_topo_feature_path),
        (epo_eeglab_path, epo_topo_feature_path),
    ],
)
def test_eeg_topoplot(file, eeglab_result_file):
    """Test eeg_topoplot feature extraction."""
    # load inst
    type_ = str(Path(file).stem)[-3:]
    inst = reader[type_](file, **kwargs[type_])
    ica = read_ica_eeglab(file)
    # get icawinv
    icawinv, _ = _retrieve_eeglab_icawinv(ica)
    # compute feature
    topo = _eeg_topoplot(inst, icawinv, ica.ch_names)
    # load from eeglab
    topo_eeglab = loadmat(eeglab_result_file)["topo"]
    # compare
    assert np.allclose(topo, topo_eeglab, equal_nan=True)


def test_eeg_topoplot_invalid_montage():
    """Test that we raise an error if the montage is badly set."""
    raw = read_raw(raw_eeglab_path, preload=True)
    ica = read_ica_eeglab(raw_eeglab_path)
    # get icawinv
    icawinv, _ = _retrieve_eeglab_icawinv(ica)
    # compute feature
    raw.info["chs"][0]["loc"] = np.array([np.nan] * 12)
    with pytest.raises(ValueError, match="Channel position for .* is missing"):
        _eeg_topoplot(raw, icawinv, ica.ch_names)
    raw.set_montage(None)
    with pytest.raises(ValueError, match="Montage is not set"):
        _eeg_topoplot(raw, icawinv, ica.ch_names)


# ----------------------------------------------------------------------------
@pytest.mark.filterwarnings("ignore:Estimated head radius.*:RuntimeWarning")
@pytest.mark.parametrize(
    ("fname", "constants_fname", "type_"),
    [
        (raw_eeglab_path, psd_constants_raw_path, "raw"),
        (epo_eeglab_path, psd_constants_epo_path, "epo"),
        (epo_long_eeglab_path, psd_constants_long_epo_path, "epo"),
    ],
)
def test_eeg_rpsd_constants(fname, constants_fname, type_):
    """Test _eeg_rpsd_constants function."""
    inst = reader[type_](fname, **kwargs[type_])
    ica = read_ica_eeglab(fname)
    ncomp, nfreqs, n_points, nyquist, index, window, subset = _eeg_rpsd_constants(
        inst, ica
    )

    constants_eeglab = loadmat(constants_fname)["constants"][0, 0]
    ncomp_eeglab = constants_eeglab["ncomp"][0, 0]
    nfreqs_eeglab = constants_eeglab["nfreqs"][0, 0]
    n_points_eeglab = constants_eeglab["n_points"][0, 0]
    nyquist_eeglab = constants_eeglab["nyquist"][0, 0]
    index_eeglab = constants_eeglab["index"]
    window_eeglab = constants_eeglab["window"]
    subset_eeglab = constants_eeglab["subset"]

    assert ncomp == ncomp_eeglab
    assert nfreqs == nfreqs_eeglab
    assert n_points == n_points_eeglab
    assert nyquist == nyquist_eeglab
    assert np.allclose(index, index_eeglab - 1)

    # window and subset are not squeezed in matlab
    assert window_eeglab.shape[0] == 1
    assert np.allclose(window, window_eeglab[0, :])

    # for subsets, compare if the same elements are in both
    assert subset_eeglab.shape[0] == 1
    assert len(set(list(subset)).difference(set(list(subset_eeglab[0, :] - 1)))) == 0
    assert len(set(list(subset_eeglab[0, :] - 1)).difference(set(list(subset)))) == 0


@pytest.mark.filterwarnings("ignore:Estimated head radius.*")
@pytest.mark.parametrize(
    ("fname", "constants_fname", "step_by_step_fname", "psd_fname", "type_"),
    [
        (
            raw_eeglab_path,
            psd_constants_raw_path,
            psd_steps_raw_path,
            psd_raw_path,
            "raw",
        ),
        (
            epo_eeglab_path,
            psd_constants_epo_path,
            psd_steps_epo_path,
            psd_epo_path,
            "epo",
        ),
        (
            epo_long_eeglab_path,
            psd_constants_long_epo_path,
            psd_steps_long_epo_path,
            psd_long_epo_path,
            "epo",
        ),
    ],
)
def test_eeg_rpsd(fname, constants_fname, step_by_step_fname, psd_fname, type_):
    """Test eeg_rpsd function that extract the PSD feature from the IC."""
    psd1 = loadmat(step_by_step_fname)["psd"]
    psd2 = loadmat(psd_fname)["psd"]
    assert np.allclose(psd1, psd2, atol=1e-4)

    # clean-up
    psd_eeglab = psd2.copy()
    del psd1
    del psd2

    # compute psd in Python
    inst = reader[type_](fname, **kwargs[type_])
    ica = read_ica_eeglab(fname)
    icaact = _compute_ica_activations(inst, ica)

    # retrieve subset from eeglab
    constants_eeglab = loadmat(constants_fname)["constants"][0, 0]
    assert constants_eeglab["subset"].shape[0] == 1
    subset_eeglab = constants_eeglab["subset"][0, :] - 1

    # retrieve the rest from python
    ncomp, nfreqs, n_points, nyquist, index, window, _ = _eeg_rpsd_constants(inst, ica)

    # compute psdmed
    psdmed = _eeg_rpsd_compute_psdmed(
        inst, icaact, ncomp, nfreqs, n_points, nyquist, index, window, subset_eeglab
    )

    # format and compare
    psd = _eeg_rpsd_format(psdmed)
    assert np.allclose(psd, psd_eeglab, atol=1e-5)


# ----------------------------------------------------------------------------
@pytest.mark.filterwarnings("ignore:Estimated head radius.*:RuntimeWarning")
def test_eeg_autocorr_welch():
    """Test eeg_autocorr_welch feature used on long raw datasets."""
    raw = read_raw(raw_eeglab_path, preload=True)
    ica = read_ica_eeglab(raw_eeglab_path)
    icaact = _compute_ica_activations(raw, ica)
    autocorr = _eeg_autocorr_welch(raw, ica, icaact)
    assert autocorr.shape[1] == 100  # check resampling
    autocorr_eeglab = loadmat(autocorr_raw_path)["autocorr"]
    assert np.allclose(autocorr, autocorr_eeglab)


@pytest.mark.filterwarnings("ignore:Estimated head radius.*:RuntimeWarning")
def test_eeg_autocorr():
    """Test eeg_autocorr feature used on short raw datasets."""
    # Raw between 1 and 5 seconds
    raw = read_raw(raw_short_eeglab_path, preload=True)
    ica = read_ica_eeglab(raw_short_eeglab_path)
    icaact = _compute_ica_activations(raw, ica)
    autocorr = _eeg_autocorr(raw, ica, icaact)
    assert autocorr.shape[1] == 100  # check resampling
    autocorr_eeglab = loadmat(autocorr_short_raw_path)["autocorr"]
    assert np.allclose(autocorr, autocorr_eeglab, atol=1e-7)

    # Raw shorter than 1 second
    raw = read_raw(raw_very_short_eeglab_path, preload=True)
    ica = read_ica_eeglab(raw_very_short_eeglab_path)
    icaact = _compute_ica_activations(raw, ica)
    autocorr = _eeg_autocorr(raw, ica, icaact)

    autocorr_eeglab = loadmat(autocorr_very_short_raw_path)["autocorr"]
    assert np.allclose(autocorr, autocorr_eeglab, atol=1e-6)


@pytest.mark.filterwarnings("ignore:Estimated head radius.*:RuntimeWarning")
def test_eeg_autocorr_fftw():
    """Test eeg_autocorr_fftw feature used on epoch datasets."""
    epochs = read_epochs_eeglab(epo_eeglab_path)
    ica = read_ica_eeglab(epo_eeglab_path)
    icaact = _compute_ica_activations(epochs, ica)
    autocorr = _eeg_autocorr_fftw(epochs, ica, icaact)
    assert autocorr.shape[1] == 100  # check resampling
    autocorr_eeglab = loadmat(autocorr_epo_path)["autocorr"]
    assert np.allclose(autocorr, autocorr_eeglab, atol=1e-7)


def test_resampling(rng):
    """Test that we correctly resample the autocorrelation feature."""
    # similar dataset shape to the MNE sample dataset
    data = rng.integers(1, 10, (15, 601))

    # with integer
    resamp = _resample(data, fs=600)
    assert resamp.shape[1] == 101

    # with floats
    for fs in np.arange(599.2, 600, 0.1):
        resamp = _resample(data, fs=fs)
        assert resamp.shape[1] == 101
    for fs in np.arange(600.1, 600.9, 0.1):
        resamp = _resample(data, fs=fs)
        assert resamp.shape[1] == 101

    # similar dataset shape to the EEGLAB sample dataset
    data = rng.integers(1, 10, (30, 129))

    # with integer
    resamp = _resample(data, fs=128)
    assert resamp.shape[1] == 101

    # with floats
    for fs in np.arange(127.2, 128, 0.1):
        resamp = _resample(data, fs=fs)
        assert resamp.shape[1] == 101
    for fs in np.arange(128.1, 128.9, 0.1):
        resamp = _resample(data, fs=fs)
        assert resamp.shape[1] == 101



from __future__ import annotations  # c.f. PEP 563, PEP 649

from typing import TYPE_CHECKING

import numpy as np
from mne import pick_info, pick_types  # Changed this line - pick_types is now in mne directly
from mne.channels.layout import _find_topomap_coords
from mne.defaults import _BORDER_DEFAULT, _EXTRAPOLATE_DEFAULT, _INTERPOLATION_DEFAULT
from mne.utils import _validate_type, check_version
from mne.viz.topomap import _check_extrapolate, _make_head_outlines, _setup_interp

if check_version("mne", "1.6"):
    def _pick_data_channels(info, exclude=()):
        return pick_types(info, meg=False, eeg=True, exclude=exclude)
    from mne._fiff.pick import _picks_to_idx
else:
    from mne.io.pick import _pick_data_channels, _picks_to_idx

from ..utils._checks import _validate_ica
from ..utils._docs import fill_doc

if TYPE_CHECKING:
    from typing import Union

    from mne import Info
    from mne.preprocessing import ICA
    from numpy.typing import NDArray


@fill_doc
def get_topomaps(
    ica: ICA,
    picks=None,
    res: int = 64,
    image_interp: str = _INTERPOLATION_DEFAULT,  # 'cubic'
    border: Union[float, str] = _BORDER_DEFAULT,  # 'mean'
    extrapolate: str = _EXTRAPOLATE_DEFAULT,  # 'auto' -> 'head' (EEG), 'local' (MEG)
) -> dict[str, NDArray[float]]:
    """Generate an array of scalp topographies for the picked components.

    Parameters
    ----------
    ica : ICA
        MNE `~mne.preprocessing.ICA` decomposition.
    picks : int | list of int | slice | None
        Indices of the independent components (ICs) to select.
        If an integer, represents the index of the IC to pick.
        Multiple ICs can be selected using a list of int or a slice.
        The indices are 0-indexed, so ``picks=1`` will pick the second IC: ``ICA001``.
        ``None`` (default) will pick all independent components in the order fitted.
    %(res_topomap)s
    %(image_interp_topomap)s
    %(border_topomap)s
    %(extrapolate_topomap)s

    Returns
    -------
    topomaps : dict of array of shape (n_components, n_pixels, n_pixels)
        Dictionary of ICs topographic maps for each channel type.
    """
    _validate_ica(ica)
    if isinstance(picks, str):
        raise TypeError(
            "Argument 'picks' should be an integer or a list of integers to select the "
            "ICs. Strings are not supported."
        )
    ic_picks = _picks_to_idx(ica.n_components_, picks)
    _validate_type(res, "int", "res", "int")
    if res <= 0:
        raise ValueError(
            f"Argument 'res' should be a strictly positive integer. Provided '{res}' "
            "is invalid."
        )
    # image_interp, border are validated by _setup_interp
    # extrapolate is validated by _check_extrapolate

    # prepare ICs
    data = np.dot(
        ica.mixing_matrix_.T,
        ica.pca_components_[: ica.n_components_],
    )
    # list channel types
    ch_picks = _pick_data_channels(ica.info, exclude=())
    ch_types = pick_info(ica.info, ch_picks).get_channel_types(unique=True)

    # compute topomaps
    topomaps = dict()
    for ch_type in ch_types:
        topomaps[ch_type] = np.zeros((ic_picks.size, res, res))
        sel = _picks_to_idx(ica.info, picks=ch_type)
        info = pick_info(ica.info, sel)
        for k, component in enumerate(ic_picks):
            topomaps[ch_type][k, :, :] = _get_topomap_array(
                data[component, sel], info, res, image_interp, border, extrapolate
            )
    return topomaps


@fill_doc
def _get_topomap_array(
    data: NDArray[float],
    info: Info,
    res: int = 64,
    image_interp: str = _INTERPOLATION_DEFAULT,  # 'cubic'
    border: Union[float, str] = _BORDER_DEFAULT,  # 'mean'
    extrapolate: str = _EXTRAPOLATE_DEFAULT,  # 'auto' -> 'head' (EEG), 'local' (MEG)
) -> NDArray[float]:
    """Generate a scalp topographic map (n_pixels, n_pixels).

    Parameters
    ----------
    data : array of shape (n_channels,)
        The data points used to generate the topographic map.
    info : Info
        Instance of `mne.Info` with the montage associated with the ``(n_channels,)``
        points.
    %(res_topomap)s
    %(image_interp_topomap)s
    %(border_topomap)s
    %(extrapolate_topomap)s

    Returns
    -------
    topomap : array of shape (n_pixels, n_pixels)
        Topographic map array.
    """
    ch_type = info.get_channel_types(unique=True)
    assert len(ch_type) == 1  # sanity-check
    ch_type = ch_type[0]
    picks = list(range(data.shape[0]))
    sphere = np.array([0.0, 0.0, 0.0, 0.095])

    # inferring (x, y) coordinates form mne.Info instance
    pos = _find_topomap_coords(info, picks=picks, sphere=sphere, ignore_overlap=True)
    extrapolate = _check_extrapolate(extrapolate, ch_type)

    # interpolation, valid only for MNE ≥ 1.1
    outlines = _make_head_outlines(sphere, pos, None, (0.0, 0.0))
    extent, Xi, Yi, interp = _setup_interp(
        pos, res, image_interp, extrapolate, outlines, border
    )
    interp.set_values(data)
    topomap = np.flipud(
        interp.set_locations(Xi, Yi)()
    )  # Zi, shape (n_pixels, n_pixels)
    np.nan_to_num(topomap, nan=0.0, copy=False)
    topomap = topomap / np.max(np.abs(topomap))  # standardize
    return topomap  # (n_pixels, n_pixels)
